"""
Contains a general base class to define 
stochastic epidemiological models in
populations of constant size.
"""

import warnings
from itertools import groupby

import numpy as np

# Try to import the original SamplableSet,
# but if that doesn't work, use the mock version
# that's implemented in this package
try:
    from SamplableSet import SamplableSet
except ModuleNotFoundError as e: # pragma: no cover
    warnings.warn("Couldn't find the efficient implementation of `SamplableSet` (see github.com/gstonge/SamplableSet). Proceeding with less efficient implementation.")
    from epipack.mock_samplable_set import MockSamplableSet as SamplableSet

from epipack.mock_samplable_set import choice as _choice

# define some integer pointers to positions in lists/tuples

# positions of compartments in event tuples
_TRANSMITTING_COMPARTMENT = 0
_AFFECTED_SOURCE_COMPARTMENT = 1
_AFFECTED_TARGET_COMPARTMENT = 2

# position of the edge weight in weighted edge tuples
_EDGE_WEIGHT = 2

# positions of items in the event lists
_EVENTS = 0
_RATES = 1
_LINK_PROCESS_INDICES = 2


class StochasticEpiModel():
    """
    A general class to define any
    compartmental epidemiological model
    that can be run in a well-mixed
    system or on a weighted, directed network.
    By default, the epidemiological process is considered
    to run in a well-mixed system.

    Parameters
    ----------
    compartments : :obj:`list` of :obj:`string`
        A list containing compartment strings.
    N_nodes : int
        The number of nodes in the system.
    edge_weight_tuples : list of tuples of (int, int, float), default = None
        Choose this ption
        The links along which transmissions can take place.

        .. code:: python

            [ (source_id, target_id, weight), ... ]

    directed : bool, default = False
        If `directed` is False, each entry in the edge_weight_tuples
        is considered to equally point from `target_id` to `source_id`
        with weight `weight`.
    well_mixed_mean_contact_number : int, default = 1
        By default, the epidemiological process is considered to run
        in a well-mixed population where for each link-transmission
        event, a node is assumed to have contact to exactly one
        other node. Increase this contact number by adjusting this
        parameter.

    Attributes
    ----------
    compartments : :obj:`list` of :obj:`string`
        A list containing strings or other hash-able types
        that describe each compartment,
        (e.g. "S", "I", etc.).
    N_comp : :obj:`int`
        Number of compartments (including population number)
    node_status : numpy.ndarray of int
        Each entry gives the compartment that the
        corresponding node is part of.

    Example
    -------

    .. code:: python

        >>> epi = StochasticEpiModel(["S","I","R"],10)
        >>> print(epi.compartments)
        [ "S", "I", "R" ]

    """

    def __init__(self,compartments,N,edge_weight_tuples=None,directed=False,well_mixed_mean_contact_number=1):
        """
        """

        # initial conditions
        self.y0 = None

        # compartment definitions
        self.compartments = list(compartments)
        self.N_comp = len(self.compartments)

        # lists defining events and their rates
        self.node_transition_events = None
        self.link_transmission_events = None
        self.node_and_link_events = None
        self.conditional_link_transmission_events = None
        self.transitioning_compartments = None
        self.transmitting_compartments = None
        self.node_status = None

        # check whether to initiate as a network or well-mixed simulation
        if edge_weight_tuples is not None:
            self.set_network(N, edge_weight_tuples, directed)
        else:
            self.set_well_mixed(N,well_mixed_mean_contact_number)

        self._simulation_ended = False

    def set_network(self,N_nodes,edge_weight_tuples,directed=False):
        """
        Define the model to run on a network.

        Parameters
        ---------
        N_nodes : int
            Number of nodes in the system
        edge_weight_tuples : list of tuple of (int, int, float)
            The links along which transmissions can take place.

                .. code:: python

                    [ (source_id, target_id, weight), ... ]

        directed : bool, default = False
            If `directed` is False, each entry in the edge_weight_tuples
            is considered to equally point from `target_id` to `source_id`
            with weight `weight`.
        """

        self.N_nodes = int(N_nodes)

        # min weights and max weights have to be set for the efficient
        # sampling within sets
        edge_weights = [ e[_EDGE_WEIGHT] for e in edge_weight_tuples ]
        if len(edge_weights) == 0:
            min_weight = 1
            max_weight = 1
        else:
            min_weight = min(edge_weights)
            max_weight = max(edge_weights)

        # construct network as list of neighbor dictionaries
        graph = [ {} for n in range(N_nodes) ]
        for source, target, weight in edge_weight_tuples:
            graph[source][target] = weight
            if not directed:
                graph[target][source] = weight

        # construct  network as list of SamplableSets, because we want to sample neighbors
        self.graph = [ SamplableSet(min_weight,max_weight,neighbors,cpp_type='int') for neighbors in graph ]

        # this model is a network model
        self.is_network_model = True

        # compute out degrees in order to compute the maximum reaction rate for link processes
        self.out_degree = np.array([ len(neighbors) for neighbors in self.graph ], dtype=int)
        self.out_strength = np.array([ neighbors.total_weight() for neighbors in self.graph ], dtype=float)

        if self.node_status is not None:
            self.set_node_statuses(self.node_status)

        return self

    def set_well_mixed(self,N_nodes,mean_contact_number):
        """
        Define the model to run in a well-mixed system.

        Parameters
        ---------
        N_nodes : int
            Number of nodes in the system
        mean_contact_number : int
            Each node is assumed to be in contact with `mean_contact_number`
            other nodes at all times. These neighbors will be sampled randomly
            from the set of all remaining nodes every time a link transmission
            event happens.
        """

        if mean_contact_number != int(mean_contact_number):
            raise TypeError("`mean_contact_number` must be of type `int`")

        # define homogeneous degree across the network
        self.N_nodes = int(N_nodes)
        self.k0 = int(mean_contact_number)

        # compute out degrees in order to compute the maximum reaction rate for link processes
        self.out_degree = np.ones((self.N_nodes,),dtype=int) * self.k0
        self.out_strength = self.out_degree

        self.is_network_model = False

        return self

    def get_compartment_id(self,C):
        """Get the integer ID of a compartment ``C``"""
        return self.compartments.index(C)

    def get_compartment(self,iC):
        """Get the compartment, given an integer ID ``iC``"""
        return self.compartments[iC]

    def set_node_transition_processes(self,process_list):
        """
        Define the linear node transition processes between compartments.

        Parameters
        ==========
        process_list : :obj:`list` of :obj:`tuple`
            A list of tuples that contains transitions rates in the following format:

            .. code:: python

                [
                    ("source_compartment", rate, "target_compartment" ),
                    ...
                ]

        Example
        =======
        For an SEIR model.

        .. code:: python

            epi.set_node_transition_processes([
                ("E", symptomatic_rate, "I" ),
                ("I", recovery_rate, "R" ),
            ])
        """

        # each compartment is associated with a list of events that it can be
        # involved in (event = first entry)
        # Each event is associated with a rate (rate = second entry)
        self.node_transition_events = [ [[],[]] for n in range(self.N_comp) ]

        # iterate through processes
        for source, rate, target in process_list:

            _s = self.get_compartment_id(source)
            _t = self.get_compartment_id(target)

            # add event and rate to the list of events for this compartment
            self.node_transition_events[_s][_EVENTS].append((-1,_s,_t))
            self.node_transition_events[_s][_RATES].append(float(rate))

        # convert to numpy arrays
        for _c0, events in enumerate(self.node_transition_events):
            self.node_transition_events[_c0][_EVENTS] = np.array(events[_EVENTS])
            self.node_transition_events[_c0][_RATES] = np.array(events[_RATES])

        # if link events have been set before, sew both events together
        # to form a unified event list for each compartment
        if self.link_transmission_events is not None:
            self._zip_events()

        return self

    def set_link_transmission_processes(self,process_list):
        r"""
        Define link transmission processes between compartments.

        Parameters
        ----------
        process_list : :obj:`list` of :obj:`tuple`
            A list of tuples that contains transitions rates in the following format:

            .. code:: python

                [
                    ("source_compartment",
                     "target_compartment_initial",
                     rate
                     "source_compartment",
                     "target_compartment_final",
                     ),
                    ...
                ]

        Example
        -------
        For an SEIR model.

        .. code:: python

            epi.set_link_transmission_processes([
                ("I", "S", +1, "I", "E" ),
            ])
        """

        # each compartment is associated with a list of events that it can be
        # involved in (event = first entry)
        # Each event is associated with a rate (rate = second entry)
        self.link_transmission_events = [ [[],[]] for n in range(self.N_comp) ]

        # iterate through processes
        for coupling0, coupling1, rate, affected0, affected1 in process_list:

            if coupling0 != affected0:
                raise ValueError("In process",
                                  coupling0, coupling1, "->", affected0, affected1,
                                  "The source (infecting) compartment", coupling0, "and first affected compartment (here: ", affected0,
                                  "), must be equal on both sides of the reaction equation but are not.",
                                  )

            _c0 = self.get_compartment_id(coupling0)
            _s = self.get_compartment_id(coupling1)
            _t = self.get_compartment_id(affected1)

            # add event and rate to the list of events for this compartment
            self.link_transmission_events[_c0][_EVENTS].append((_c0,_s,_t))
            self.link_transmission_events[_c0][_RATES].append(float(rate))

        # convert to numpy arrays
        for _c0, events in enumerate(self.link_transmission_events):
            self.link_transmission_events[_c0][_EVENTS] = np.array(events[_EVENTS])
            self.link_transmission_events[_c0][_RATES] = np.array(events[_RATES])

        # if node events have been set before, sew both events together
        # to form a unified event list for each compartment
        if self.node_transition_events is not None:
            self._zip_events()

        return self

    def set_conditional_link_transmission_processes(self,process_dict):
        r"""
        Define link transmission processes between compartments.

        Parameters
        ----------
        process_dict : :obj:`list` of :obj:`tuple`
            A dictionary of tuples that contains conditional transmission events in the following format:

            .. code:: python

                {
                    ( "source_base", "->", "target_base" ): [
                        ("target_base",
                         "target_compartment_initial",
                         probability
                         "target_base",
                         "target_compartment_final",
                         ),
                    ...
                   ],
                    ( "infecting", "source_base", "->", "infecting", "target_base" ): [
                        ("target_base",
                         "target_compartment_initial",
                         "->"
                         "target_base",
                         "target_compartment_final",
                         ),
                    ...
                   ]
                }

        Example
        -------
        When an `I`-node recovers (to `R`), scan all of the newly-recovered node's
        neighbors. If the neighbor is an `S`, transition the neighbor
        to `X`. If the neighbor is an `I`, transition the neighbor to `Q` with
        probability :math:`p` (implying that nothing happens to this neighbor
        with probability :math:`1-p`.

        .. code:: python

            epi.set_conditional_link_transmission_processes({
                ( "I", "->", "R" ) : [
                    ("R", "S", "->", "R", "X" ),
                    ("R", "I", p, "R", "Q" ),
                ]
            })
        """

        # each compartment is associated with a list of events that it can be
        # involved in (event = first entry)
        # Each event is associated with a rate (rate = second entry)
        self.conditional_link_transmission_events = {}

        # iterate through processes
        for triggering_event, triggered_events in process_dict.items():

            if len(triggering_event) == 5:
                coupling0, coupling1, _, affected0, affected1 = triggering_event
                if coupling0 != affected0:
                    raise ValueError("In process " +\
                                      str((coupling0, coupling1, "->", affected0, affected1)) +\
                                      " The source (infecting) compartment " + str(coupling0) +\
                                      " and first affected compartment (here: " + str(affected0) +\
                                      "), must be equal on both sides of the reaction equation but are not."
                                      )
                _c0 = self.get_compartment_id(coupling0)
                _s = self.get_compartment_id(coupling1)
                _t = self.get_compartment_id(affected1)
            elif len(triggering_event) == 3:
                source, _, target = triggering_event
                _c0 = -1
                _s = self.get_compartment_id(source)
                _t = self.get_compartment_id(target)
            else:
                raise ValueError("Only node transition or link transmission events are "
                                 "allowed to trigger conditional link transmission events "
                                 '(invalid event: "'+ str(triggering_event) +'"')

            event = (_c0, _s, _t)
            self.conditional_link_transmission_events[event] = {}

            # check triggered events for correct format
            for triggered_event in triggered_events:

                if len(triggered_event) != 5:
                    raise ValueError("Only link transmission events are allowed to be triggered (invalid event:", triggered_event)

                coupling0, coupling1, probability, affected0, affected1 = triggered_event

                if coupling0 != affected0:
                    raise ValueError("In process",
                                      coupling0, coupling1, "->", affected0, affected1,
                                      "The source (infecting) compartment", coupling0, "and first affected compartment (here: ", affected0,
                                      "), must be equal on both sides of the reaction equation but are not.",
                                      )

                if self.get_compartment_id(coupling0) != event[_AFFECTED_TARGET_COMPARTMENT]:
                    raise ValueError("In conditional link transmission process",
                                      coupling0, coupling1, "->", affected0, affected1,
                                      "The source (infecting) compartment", coupling0, "must be equal to the affected (target) compartment of",
                                      "the triggering event", triggering_event, " but is not."
                                      )


            # group the triggered events by the same source compartment.
            triggered_events = sorted(triggered_events,key=lambda x: x[1])
            for source, triggered_event_group in groupby(triggered_events, key=lambda x: x[1]):
                probabilities = []
                these_events = []

                for coupling0, coupling1, probability, affected0, affected1 in triggered_event_group:

                    try:
                        probability = float(probability)
                    except ValueError as e:
                        probability = 1.0

                    _c0 = self.get_compartment_id(coupling0)
                    _s = self.get_compartment_id(coupling1)
                    _t = self.get_compartment_id(affected1)

                    this_triggered_event = (_c0, _s, _t)
                    these_events.append(this_triggered_event)
                    probabilities.append(probability)

                sum_p = sum(probabilities)
                if sum_p - 1.0 > 1e-14:
                    raise ValueError("probabilities sum to values larger than one for the event group above.")
                elif sum_p < 1.0:
                    probabilities.append(1-sum_p)
                    these_events.append((_c0, _s, _s))

                probabilities = np.array(probabilities,dtype=float)
                probabilities /= probabilities.sum()

                self.conditional_link_transmission_events[event][_s] = ( probabilities, these_events )

        return self

    def reactions_may_still_occur(self,state):
        """
        Reactions can only take place if there's still nodes
        that can transmit AND nodes that can be
        infected or spontaneously change state.
        """
        if len(self.transmitting_compartments) > 0:
            return state[self.transitioning_compartments].sum() > 0 or\
                (state[self.transmitting_compartments[:,0]] * state[self.transmitting_compartments[:,1]]).sum() > 0
        else:
            return state[self.transitioning_compartments].sum() > 0

    def _zip_events(self):
        """
        Concatenate link and node events.
        """
        self.node_and_link_events = [ None for n in range(self.N_comp) ]
        transitioning_compartments = set()
        transmitting_compartments = set()

        # for each compartment, concatenate node and link event lists
        for comp in range(self.N_comp):
            link_events = self.link_transmission_events[comp]
            node_events = self.node_transition_events[comp]

            # the last entry in this list saves the event index range that are link events.
            # we need this, because for every node, these rates have to be multiplied by
            # the out degree later on.
            if len(node_events[_EVENTS]) == 0 and len(link_events[_EVENTS]) == 0:
                events = np.array([])
                rates = np.array([])
                indices = [0, 0]
            elif len(node_events[_EVENTS]) != 0 and len(link_events[_EVENTS]) == 0:
                events = node_events[_EVENTS]
                rates = node_events[_RATES]
                indices = [0, 0]
            elif len(node_events[_EVENTS]) == 0 and len(link_events[_EVENTS]) != 0:
                events = link_events[_EVENTS]
                rates = link_events[_RATES]
                indices = [0, len(link_events[_EVENTS])]
            else:
                events = np.concatenate((node_events[_EVENTS],link_events[_EVENTS]))
                rates = np.concatenate((node_events[_RATES],link_events[_RATES]))
                indices = [len(node_events[_RATES]), len(node_events[_RATES]) + len(link_events[_RATES])]

            self.node_and_link_events[comp] = [ events, rates, indices ]

            # save the compartments that can change their status.
            # This is important because it might happen that nodes
            # have a non-zero reaction rate but no targets left
            # at which point the simulation does never halt.
            for infecting, transitioning, _ in events:
                if infecting < 0:
                    transitioning_compartments.add(transitioning)
                else:
                    transmitting_compartments.add((infecting,transitioning))

        self.transitioning_compartments = np.array(list(transitioning_compartments),dtype=int)
        self.transmitting_compartments = np.array(list(transmitting_compartments),dtype=int)

    def set_random_initial_conditions(self, initial_conditions):
        """
        Set random initial conditions for each compartment.

        Parameters
        ----------
        initial_conditions : dict
            A dictionary that maps a compartment to a number of nodes
            that should be sampled to be in this compartment initially.
            Unset compartmens are assumed to have an initial condition of zero.
        """

        if type(initial_conditions) == dict:
            initial_conditions = list(initial_conditions.items())

        # initially, all nodes can be distributed
        nodes_left = set(range(self.N_nodes))

        # save node status and initial conditions
        node_status = np.zeros(self.N_nodes,dtype=int)
        y0 = np.zeros(self.N_comp)
        total = 0

        # iterate through all initial conditions
        for compartment, amount in initial_conditions:

            # ignore nullified initial conditions
            if amount == 0:
                continue

            # count the number of nodes in this compartment
            total += amount
            comp_id = self.get_compartment_id(compartment)

            # check whether it was defined before
            if y0[comp_id] != 0:
                raise ValueError('Double entry in initial conditions for compartment '+str(compartment))
            else:
                # if node, sample `amount` nodes from the remaining set,
                # take those nodes out of the set and carry on with the set of remaining nodes
                these_nodes = np.random.choice(list(nodes_left),size=amount,replace=False)
                y0[comp_id] = amount
                node_status[these_nodes] = comp_id
                nodes_left = nodes_left.difference(these_nodes)

        # check that the initial conditions contain all nodes
        if np.abs(total-self.N_nodes)/self.N_nodes > 1e-14:
            raise ValueError('Sum of initial conditions does not equal unity.')

        self.set_node_statuses(node_status)
        self.y0 = y0

        return self

    def set_node_statuses(self,node_status):
        """
        Set all node statuses at once and evaluate events and rates accordingly.
        Can be used to set initial conditions.

        Parameters
        ----------
        node_status : numpy.ndarray of int
            For each node, this array contains the node's compartment index.
        """

        if len(node_status) != self.N_nodes:
            raise ValueError("`node_status` must carry N_nodes values")

        if self.link_transmission_events is None:
            self.set_link_transmission_processes([])
        if self.node_transition_events is None:
            self.set_node_transition_processes([])

        # filter out the minimally possible rate (other than zero) and the maximally
        # possible rate 
        smax = self.out_strength.max()
        smin = self.out_strength.min()
        compartment_mins = np.zeros(self.N_comp)
        compartment_maxs = np.zeros(self.N_comp)
        for compartment, (event, rates) in enumerate(self.node_transition_events):
            for rate in rates:
                compartment_mins[compartment] += rate
                compartment_maxs[compartment] += rate

        for compartment, (event, rates) in enumerate(self.link_transmission_events):
            for rate in rates:
                compartment_mins[compartment] += rate * smin
                compartment_maxs[compartment] += rate * smax

        if np.all(compartment_mins==0):
            if smin == 0 and smax == 0:
                # in case of empty network,
                # set some default values.
                # This is okay because
                # the event set will still be empty
                compartment_min = 1e-128
                compartment_max = 1e128
        else:
            compartment_min = (compartment_mins[compartment_mins>0]).min()
            compartment_max = (compartment_maxs).max()

        # set these node statuses
        self.node_status = node_status.copy()

        # construct a SamplableSet in which we insert the node reaction rates
        self.all_node_events = SamplableSet(compartment_min,compartment_max,cpp_type='int')

        # construct a list that contains events and rates for each node once it was triggered
        self.node_event_probabilities = [ [] for n in range(self.N_nodes) ]

        # set each node status according to the passed node_status array
        for node, status in enumerate(self.node_status):
            self.set_node_status(node, status)

        y0 = np.zeros(self.N_comp)
        for c in range(self.N_comp):
            y0[c] += np.count_nonzero(self.node_status == c)

        self.y0 = y0

        self._simulation_ended = False

        return self

    def set_node_status(self,node,status):
        """
        Set the status of node `node` to `status`

        Parameters
        ----------
        node : int
            The index of the node
        status : int
            The index of the compartment that the node changes into.
        """

        # get the events that can happen to a node of this status/compartment
        status_events = self.node_and_link_events[status]

        # if no events can happen, remove the node from 
        # the event set and set the node events to an empty set
        if len(status_events[_RATES]) == 0:
            del self.all_node_events[node]
            self.node_event_probabilities[node] = []
        else:
            # otherwise, get a copy of this compartment's rate vector
            these_rates = status_events[_RATES].copy()
            # scale the link events by this node's out degree            
            scale_from, scale_to = status_events[_LINK_PROCESS_INDICES]
            these_rates[scale_from:scale_to] *= self.out_strength[node]
            total_rate = these_rates.sum()
            # add the node to the node event set, IF it can react
            # after all (it may not be able to react if it has no
            # neighbors).
            if total_rate > 0:
                self.all_node_events[node] = total_rate
                # save the probabilities with which each of the node's events happens
                self.node_event_probabilities[node] = these_rates / total_rate

        # finally, set node status
        self.node_status[node] = status

        return self

    def get_total_event_rate(self):
        """
        Get the total event rate.
        """
        # sometimes, it happens after many deletions and inserts,
        # numerical errors add up and the set has non-zero total weight
        # but is not empty. We have to catch these events.
        if len(self.all_node_events) > 0:
            return self.all_node_events.total_weight()
        else:
            self.all_node_events.clear()
            return 0.0


    def get_true_total_event_rate(self):
        """
        Get the true total event rate.
        """

        total_rate = 0.0

        for node, max_rate in self.all_node_events:
            status = self.node_status[node]

            node_and_link_events = self.node_and_link_events[status]
            events = node_and_link_events[_EVENTS]
            rates = node_and_link_events[_RATES]
            l0, l1 = node_and_link_events[_LINK_PROCESS_INDICES]

            has_link_processes = l1 > 0

            if not has_link_processes:
                total_rate += max_rate
            else:
                total_rate += rates[0:l0].sum()
                coupling_compartments = np.array([ e[_AFFECTED_SOURCE_COMPARTMENT] for e in events[l0:l1] ])
                these_rates = rates[l0:l1]

                if self.is_network_model:
                    neighs = self.graph[node]

                    for n in neighs:

                        try:
                            n = n[0]
                        except (IndexError, TypeError) as e:
                            pass

                        neigh_status = self.node_status[n]

                        ndx = np.where(coupling_compartments == neigh_status)[0]
                        total_rate += these_rates[ndx].sum()
                else:
                    for C in coupling_compartments:
                        nodes_of_this_compartment = self.y0[C]
                        total_rate += these_rates[C] * nodes_of_this_compartment * self.k0 / (self.N_nodes-1)

        if total_rate == 0.0:
            self.set_simulation_has_ended()

        return total_rate

    def simulation_has_ended(self):
        """
        Check wether the simulation can be continued in its current state.
        """
        return self._simulation_ended

    def set_simulation_has_ended(self):
        self._simulation_ended = True

    def set_simulation_has_not_ended(self):
        self._simulation_ended = False

    def get_time_leap(self):
        """
        Sample a time leap from an exponential distribution
        according to the current total event rate
        """
        return np.random.exponential(1/self.get_total_event_rate())

    def get_compartment_changes(self):
        """
        Let an event take place according to the new time and return the change in compartment counts.
        """
        reacting_node = self.get_reacting_node()
        return self.make_random_node_event(reacting_node)

    def get_reacting_node(self):
        """
        Get a reacting node with probability proportional to its reaction rate.
        """
        return self.all_node_events.sample()[0]

    def make_random_node_event(self, reacting_node):
        """
        Let a random node event happen according to the event probabilities
        of this node's status.

        Parameters
        ----------
        reacting_node : int
            the index of the node that reacts

        Returns
        -------
        compartment_changes : list of tuples of int
            Each tuple contains the index of
            the compartment that loses a member on the first position
            and the index of the compartment that gains a member.
        """

        # once a node that reacts, obtain the probabilities with which
        # each of its events happens, and its current status
        status = self.node_status[reacting_node]
        node_event_probabilities = self.node_event_probabilities[reacting_node]

        # sample the index of the event that happens according to the node probabilities
        # and get the corresponding event
        event_index = _choice(len(node_event_probabilities),p=node_event_probabilities)
        event = self.node_and_link_events[status][_EVENTS][event_index]

        return self.make_node_event(reacting_node, tuple(event))

    def make_node_event(self, reacting_node, event, neighbor=None):
        """
        Let a specific node event happen

        Parameters
        ----------
        reacting_node : int
            the index of the node that reacts
        event : tuple of int
            three-entry long tuple that characterizes this event
        neighbor : int, default = None
            specify the neighbor to which this specific event happens

        Returns
        -------
        compartment_changes : list of tuples of int
            Each tuple contains the index of
            the compartment that loses a member on the first position
            and the index of the compartment that gains a member.
        """

        # classify the event and changing compartments
        is_transmission_event = event[_TRANSMITTING_COMPARTMENT] != -1
        losing_compartment = event[_AFFECTED_SOURCE_COMPARTMENT]
        gaining_compartment = event[_AFFECTED_TARGET_COMPARTMENT]

        compartment_changes = []

        # if is transmission event
        if is_transmission_event:
            if neighbor is None:
                # if no reacting neighbor was specified    
                if self.is_network_model:
                    # sample a neighbor from the reacting node's neighbor set
                    neighbor, _ = self.graph[reacting_node].sample()
                else:
                    # in well-mixed, sample any other node
                    neighbor = np.random.randint(self.N_nodes-1)
                    if neighbor >= reacting_node:
                        neighbor += 1
            # if the sampled neighbor is of the affected compartment,
            # the affected node is the neighbor
            if self.node_status[neighbor] == event[_AFFECTED_SOURCE_COMPARTMENT]:
                affected_node = neighbor
            else:
                # otherwise, nothing happens
                return compartment_changes
        else:
            # if it's a node transition event, this node is affected
            affected_node = reacting_node

        new_compartment = event[_AFFECTED_TARGET_COMPARTMENT]
        self.set_node_status(affected_node,new_compartment)

        # check for conditional processes based on this process
        if self.conditional_link_transmission_events is not None:

            # check whether an entry exists for this event
            try:
                conditional_events = self.conditional_link_transmission_events[event]
                has_conditional_transmissions = True
            except KeyError as e:
                has_conditional_transmissions = False

            # if it exists
            if has_conditional_transmissions:

                # sample neighbors of the affected node (might include this reacting node)
                if self.is_network_model:
                    neighbors = self.graph[affected_node]
                else:
                    neighbors = np.random.choice(self.N_nodes-1,size=int(self.out_degree[affected_node]),replace=False)
                    neighbors[neighbors >= affected_node] += 1

                # iterate through neighbors
                for n in neighbors:
                    # samplableset returns node ids together with weight, therefore we 
                    # have to ask for the first entry of n
                    try:
                        n = n[0]
                    except IndexError as e:
                        pass

                    # if the compartment of the neighbor is affected by a conditional event,
                    # let this conditional event happen to the neighbor, as triggered
                    # by the affected node
                    try:
                        probabilities, conditional_transmission_events = conditional_events[self.node_status[n]]
                    except KeyError as e:
                        continue

                    # check whether the list of proposed conditional reactions for this event
                    # is longer than one. if not, let this single event take place with the given probability
                    if len(probabilities) == 1 and (probabilities[0] == 1.0 or probabilities[0] > np.random.random()):
                        these_compartment_changes = self.make_node_event(affected_node, conditional_transmission_events[0], neighbor = n)
                        compartment_changes.extend(these_compartment_changes)
                    else:
                        # if there's a list of proposed events, each associated with a probability,
                        # choose from the list uniform at random
                        event_index = _choice(len(probabilities),p=probabilities)
                        conditional_transmission_event = conditional_transmission_events[event_index]

                        # check wether this is an event where nothing happens
                        if conditional_transmission_event[_AFFECTED_SOURCE_COMPARTMENT] != \
                                conditional_transmission_event[_AFFECTED_TARGET_COMPARTMENT]:
                            # if something happens, let it happen
                            these_compartment_changes = self.make_node_event(affected_node, conditional_transmission_event, neighbor = n)
                            compartment_changes.extend(these_compartment_changes)

        compartment_changes.append((losing_compartment, gaining_compartment))

        return compartment_changes

    def simulate(self,
                 tmax,
                 return_compartments=None,
                 sampling_dt=None,
                 max_unsuccessful=None,
                 sampling_callback=None,
                 t0=0.0,
                 stop_simulation_on_empty_network=True,
                 custom_stop_condition=None,
                 **kwargs):
        """
        Returns values of the given compartments at the demanded
        time points (as a numpy.ndarray of shape 
        ``(return_compartments), len(time_points)``.
        If ``return_compartments`` is None, all compartments will
        be returned.

        Parameters
        ----------
        tmax : float
            maximum length of the simulation
        return_compartments : list of compartments, default = None:
            The compartments for which to return time series.
            If ``None``, all compartments will be returned.
        sampling_dt : float, default = None
            Temporal distance between samples of the compartment counts.
            If ``None``, every change will be returned.
        max_unsuccessful : int, default = None
            The number of unsuccessful events after which the
            true total event rate will be evaluated (it might happen
            that a network becomes effectively disconnected while
            nodes are still associated with a maximum event rate).
            If ``None``, this number will be set equal to the number of nodes.
        sampling_callback : funtion, default = None
            A function that's called when a sample is taken

        Returns
        -------
        t : numpy.ndarray
            times at which compartment counts have been sampled
        result : dict
            Dictionary mapping a compartment to a time series of its count.
        """

        if return_compartments is None:
            return_compartments = self.compartments

        if sampling_callback is not None and sampling_dt is None:
            raise ValueError('A sampling callback function can only be set if sampling_dt is set, as well.')

        ndx = [self.get_compartment_id(C) for C in return_compartments]
        current_state = self.y0.copy()
        compartments = [ current_state.copy() ]

        t = t0
        time = [t0]

        if max_unsuccessful is None:
            max_unsuccessful = self.N_nodes
            unsuccessful = 0

        total_event_rate = self.get_total_event_rate()

        if sampling_callback is not None:
            sampling_callback()

        custom_stop_condition_triggered = False

        # Check for a) zero event rate and b) zero possibility for any nodes being changed still.
        # This is important because it might happen that nodes
        # have a non-zero reaction rate but no targets left
        # at which point the simulation will never halt.
        while (t < tmax) and \
              (not self.simulation_has_ended()) and\
              (total_event_rate > 0) and \
              (self.reactions_may_still_occur(current_state)) and\
              (not custom_stop_condition_triggered):

            # sample and advance time according to current total rate
            tau = self.get_time_leap()
            new_t = t+tau

            # break if simulation time is reached
            if new_t >= tmax:
                break

            # sample a reacting node from the reaction set
            # let the event take place and get all the compartments
            # that are associated with changes
            changing_compartments = self.get_compartment_changes()

            # only save compartment counts if anything changed
            if len(changing_compartments) > 0:

                if sampling_dt is not None:
                    # sample all the time steps that were demanded in between the two events
                    last_sample_dt = time[-1]
                    for idt in range(1,int(np.ceil((new_t-last_sample_dt)/sampling_dt))):
                        time.append(last_sample_dt+idt*sampling_dt)
                        compartments.append(current_state.copy())
                        if sampling_callback is not None:
                            sampling_callback()

                # write losses and gains into the current state vector
                for losing, gaining in changing_compartments:
                    current_state[losing] -= 1
                    current_state[gaining] += 1

                # save the current state if sampling_dt wasn't specified
                if sampling_dt is None:
                    time.append(new_t)
                    compartments.append(current_state.copy())

                unsuccessful = 0
                total_event_rate = self.get_total_event_rate()

                # save current state
                self.y0 = current_state.copy()
            else:
                unsuccessful += 1
                if unsuccessful > max_unsuccessful:
                    total_event_rate = self.get_true_total_event_rate()
                    unsuccessful = 0
                else:
                    total_event_rate = self.get_total_event_rate()


            # advance time
            t = new_t

            if custom_stop_condition is not None:
                custom_stop_condition_triggered = custom_stop_condition(current_state,self.node_status)



        if sampling_dt is not None:
            next_sample = time[-1] + sampling_dt
            if next_sample <= tmax:
                time.append(next_sample)
                compartments.append(current_state)
                if sampling_callback is not None:
                    sampling_callback()


        # convert to result dictionary
        time = np.array(time)
        result = np.array(compartments)

        if sampling_callback is not None:
            sampling_callback()

        if not self.reactions_may_still_occur(current_state) or\
           (total_event_rate == 0.0 and stop_simulation_on_empty_network):
            self.set_simulation_has_ended()

        return time, { compartment: result[:,c_ndx] for c_ndx, compartment in zip(ndx, return_compartments) }

class StochasticSIModel(StochasticEpiModel):
    r"""
    An SI model derived from :class:`epipack.stochastic_epi_models.StochasticEpiModel`.

    Parameters
    ----------
    N : int
        Number of nodes.
    infection_rate : float
        Inverse mean duration of a single SI-contact until infection :math:`\eta=1/\tau`.
    *args : list
        List of arguments that will be passed to 
        :class:`epipack.stochastic_epi_models.StochasticEpiModel`.
    **kwargs : dict
        Additional keyword arguments that will be passed to 
        :class:`epipack.stochastic_epi_models.StochasticEpiModel`.

    Attributes
    ----------
    compartments : list of str
        
        .. code:: python

            [ "S", "I" ]
    """

    def __init__(self, N, infection_rate, *args, **kwargs):

        StochasticEpiModel.__init__(self, list("SI"), N, *args, **kwargs)

        self.set_link_transmission_processes([
                ("I", "S", infection_rate, "I", "I"),
            ])

class StochasticSIRModel(StochasticEpiModel):
    r"""
    An SIR model derived from :class:`epipack.stochastic_epi_models.StochasticEpiModel`.

    Parameters
    ----------
    N : int
        Number of nodes.
    R0 : float
        Basic reproduction number :math:`R_0 = \eta\left< k\right>/\rho`.
    recovery_rate : float
        Inverse duration of infection :math:`\rho=1/\tau_I`.
    *args : list
        List of arguments that will be passed to 
        :class:`epipack.stochastic_epi_models.StochasticEpiModel`.
    **kwargs : dict
        Additional keyword arguments that will be passed to 
        :class:`epipack.stochastic_epi_models.StochasticEpiModel`.

    Attributes
    ----------
    compartments : list of str
        
        .. code:: python

            [ "S", "I", "R" ]
    """

    def __init__(self, N, R0, recovery_rate, *args, **kwargs):

        StochasticEpiModel.__init__(self, list("SIR"), N, *args, **kwargs)

        k0 = self.out_strength.mean() 
        infection_rate = R0 * recovery_rate / k0

        self.set_node_transition_processes([
                ("I", recovery_rate, "R"),
            ])

        self.set_link_transmission_processes([
                ("I", "S", infection_rate, "I", "I"),
            ])

class StochasticSISModel(StochasticEpiModel):
    r"""
    An SIS model derived from :class:`epipack.stochastic_epi_models.StochasticEpiModel`.

    Parameters
    ----------
    N : int
        Number of nodes.
    R0 : float
        Basic reproduction number :math:`R_0 = \eta\left< k\right>/\rho`.
    recovery_rate : float
        Inverse duration of infection :math:`\rho=1/\tau_I`.
    *args : list
        List of arguments that will be passed to 
        :class:`epipack.stochastic_epi_models.StochasticEpiModel`.
    **kwargs : dict
        Additional keyword arguments that will be passed to 
        :class:`epipack.stochastic_epi_models.StochasticEpiModel`.

    Attributes
    ----------
    compartments : list of str
        
        .. code:: python

            [ "S", "I" ]

    """

    def __init__(self, N, R0, recovery_rate, *args, **kwargs):

        StochasticEpiModel.__init__(self, list("SI"), N, *args, **kwargs)

        k0 = self.out_strength.mean() 
        infection_rate = R0 * recovery_rate / k0

        self.set_node_transition_processes([
                ("I", recovery_rate, "S"),
            ])

        self.set_link_transmission_processes([
                ("I", "S", infection_rate, "I", "I"),
            ])

if __name__ == "__main__":     # pragma: no cover

    model = StochasticEpiModel(list("SI"),3)

    model.set_node_transition_processes([("I",1.0,"S")])
    model.set_link_transmission_processes([("I","S",2.0,"I","I")])

    print(model.link_transmission_events)
    print(model.node_transition_events)
    for comp, comp_events in enumerate(model.node_and_link_events):
        print(model.get_compartment(comp), comp_events)
    
    #model.set_well_mixed(3,mean_contact_number=1)
    model.set_random_initial_conditions({"S": 2, "I": 1})

    print(model.y0)
    print(model.node_status)

    reacting_node = model.get_reacting_node()

    print("reacting node:", reacting_node)

    model.make_random_node_event(reacting_node)

    print(model.node_status)


    print("===================")


    N = 10000
    i0 = 0.05
    I0 = int(i0*N)
    S0 = N - I0

    R0 = 2.0
    mu = 1.0
    k0 = 100

    model = StochasticEpiModel(list("SI"),N,well_mixed_mean_contact_number=k0)

    model.set_node_transition_processes([("I",mu,"S")],)
    model.set_link_transmission_processes([("I","S",R0/k0,"I","I")])
    model.set_random_initial_conditions({"S":S0,"I":I0})

    t, result = model.simulate(4,sampling_dt = 0.1)

    from bfmplot import pl
    for comp, series in result.items():
        pl.plot(t, series, label=comp)

    from metapop.epi import SISModel

    model = SISModel(R0=R0,recovery_rate=mu,population_size=N)
    model.set_initial_conditions({"S":S0,"I":I0})
    result = model.get_result_dict(t)

    for comp, series in result.items():
        pl.plot(t, series, label=comp)

    pl.legend()
    print("===================")


    N = 10000
    i0 = 0.05
    I0 = int(i0*N)
    S0 = N - I0

    R0 = 2.0
    mu = 1.0
    k0 = 100

    model = StochasticEpiModel(list("SIR"),N,well_mixed_mean_contact_number=k0)

    model.set_node_transition_processes([
                                            ("I",mu,"R"),
                                        ])
    model.set_link_transmission_processes([("I","S",R0/k0*mu,"I","I")])
    model.set_random_initial_conditions({"S":S0,"I":I0})

    t, result = model.simulate(4,sampling_dt = 0.1)

    pl.figure()

    for comp, series in result.items():
        pl.plot(t, series, label=comp)

    from metapop.epi import SIRModel

    model = SIRModel(R0=R0,recovery_rate=mu,population_size=N)
    model.set_initial_conditions({"S":S0,"I":I0})
    result = model.get_result_dict(t)

    for comp, series in result.items():
        pl.plot(t, series, label=comp)
    
    pl.legend()

    print("====== BARABASI-ALBERT ========")


    N = 1000
    i0 = 0.05
    I0 = int(i0*N)
    S0 = N - I0

    R0 = 2.0
    mu = 1.0
    k0 = 100

    import networkx as nx
    G = nx.barabasi_albert_graph(N,k0)
    weighted_edge_tuples = [ (e[0], e[1], 1.0) for e in G.edges() ]
    k_norm = 2*len(weighted_edge_tuples) / N

    model = StochasticEpiModel(list("SIR"),N,edge_weight_tuples=weighted_edge_tuples)

    model.set_node_transition_processes([
                                            ("I",mu,"R"),
                                        ])
    model.set_link_transmission_processes([("I","S",R0/k_norm*mu,"I","I")])

    model.set_network(N,weighted_edge_tuples)
    model.set_random_initial_conditions({"S":S0,"I":I0})

    from time import time
    start = time()
    t, result = model.simulate(100)
    end = time()
    print("composition rejection needed", end-start,"seconds") 


    pl.figure()

    for comp, series in result.items():
        pl.plot(t, series, label=comp)

    from epipack.tests.SIR_weighted import SIR_weighted

    start = time()
    model = SIR_weighted(N,weighted_edge_tuples,infection_rate=R0/k_norm*mu,recovery_rate=mu,number_of_initially_infected=I0)
    end = time()
    print("force of infection needed", end-start,"seconds") 
    _t, _I, _R = model.simulation(100)

    pl.plot(_t, _I, label="I")
    pl.plot(_t, _R, label="R")
    

    pl.legend()
    pl.show()




"""
Visualizations of simulations with pyglet.
"""

from copy import deepcopy
from itertools import chain

import numpy as np

import pyglet
from pyglet import shapes
from pyglet.window import key, mouse, Window
from pyglet.gl import *

from epipack.networks import get_random_layout, get_grid_layout
from epipack import colors as col
from epipack.colors import colors

_colors = list(colors.values())


class SimulationStatus():
    """
    Saves information about the current simulation.

    Parameters
    ==========
    N : int
        Number of nodes
    sampling_dt : float
        The amount of simulation time that's supposed
        to pass during a single update

    Attributes
    ==========
    old_node_status : numpy.ndarray
        An array containing node statuses of the previous update
    sampling_dt : float
        The amount of simulation time that's supposed
        to pass during a single update
    simulation_ended : bool
        Whether or not the simulation is over
    paused : bool
        Whether or not the simulation is paused
    """

    def __init__(self,N,sampling_dt):
        self.old_node_status = -1e300*np.ones((N,))
        self.simulation_ended = False
        self.sampling_dt = sampling_dt
        self.paused = False

    def update(self,old_node_status):
        """
        Update the nodes statuses.
        """
        self.old_node_status = np.array(old_node_status)

    def set_simulation_status(self,simulation_ended):
        """
        Trigger the simulation to be over.
        """
        self.simulation_ended = simulation_ended


class App(pyglet.window.Window):
    """

    A pyglet Window class that makes zooming and panning convenient
    and tracks user input.

    Adapted from Peter Varo's solution
    at https://stackoverflow.com/a/19453006/4177832

    Parameters
    ==========
    width : float
        Width of the app window
    height : float
        Height of the app window
    simulation_status : SimulationStatus
        An object that tracks the simulation. Here,
        it's used to pause or increase the simulation speed.
    """

    def __init__(self, width, height, simulation_status, *args, **kwargs):
        #conf = Config(sample_buffers=1,
        #              samples=4,
        #              depth_size=16,
        #              double_buffer=True)
        self.left   = 0
        self.right  = width
        self.bottom = 0
        self.top    = height
        super().__init__(width, height, *args, **kwargs)
        self.batches = []
        self.batch_funcs = []

        #Initialize camera values
        self.left   = 0
        self.right  = width
        self.bottom = 0
        self.top    = height
        self.zoom_level = 1
        self.zoomed_width  = width
        self.zoomed_height = height

        # Set window values
        self.width  = width
        self.height = height

        self.orig_left = self.left
        self.orig_right = self.right
        self.orig_bottom = self.bottom
        self.orig_top = self.top
        self.orig_zoom_level = self.zoom_level
        self.orig_zoomed_width = self.zoomed_width
        self.orig_zoomed_height = self.zoomed_height

        self.simulation_status = simulation_status


    def add_batch(self,batch,prefunc=None):
        """
        Add a batch that needs to be drawn.
        Optionally, also pass a function that's
        triggered before this batch is drawn.
        """
        self.batches.append(batch)
        self.batch_funcs.append(prefunc)


    def on_draw(self):
        """
        Clear and draw all batches
        """
        self.clear()

        for batch, func in zip(self.batches, self.batch_funcs):
            if func is not None:
                func()
            batch.draw()

    def init_gl(self, width, height):
        # Set viewport
        glMatrixMode( GL_PROJECTION )
        glLoadIdentity()
        #try:
        glOrtho( self.left, self.right, self.bottom, self.top, 1, -1 )
        #except AttributeError as e:
        #    print(self.name) 

    def on_resize(self, width, height):
        """Rescale."""
        # Set window values
        self.width  = width
        self.height = height

        # Initialize OpenGL context
        self.init_gl(width, height)

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        """Pan."""
        # Move camera
        self.left   -= dx*self.zoom_level
        self.right  -= dx*self.zoom_level
        self.bottom -= dy*self.zoom_level
        self.top    -= dy*self.zoom_level

        glMatrixMode( GL_PROJECTION )
        glLoadIdentity()
        glOrtho( self.left, self.right, self.bottom, self.top, 1, -1 )

    def on_mouse_scroll(self, x, y, dx, dy):
        """Zoom."""
        # Get scale factor
        f = 1+dy/50
        # If zoom_level is in the proper range
        if .02 < self.zoom_level*f < 10:

            self.zoom_level *= f

            mouse_x = x/self.width
            mouse_y = y/self.height

            mouse_x_in_world = self.left   + mouse_x*self.zoomed_width
            mouse_y_in_world = self.bottom + mouse_y*self.zoomed_height

            self.zoomed_width  *= f
            self.zoomed_height *= f

            self.left   = mouse_x_in_world - mouse_x*self.zoomed_width
            self.right  = mouse_x_in_world + (1 - mouse_x)*self.zoomed_width
            self.bottom = mouse_y_in_world - mouse_y*self.zoomed_height
            self.top    = mouse_y_in_world + (1 - mouse_y)*self.zoomed_height

        glMatrixMode( GL_PROJECTION )
        glLoadIdentity()
        glOrtho( self.left, self.right, self.bottom, self.top, 1, -1 )


    def on_key_press(self, symbol, modifiers):
        """
        Check for keyboard input.
        Current inputs:

        - backspace or CMD+0: reset view
        - up : increase simulation speed
        - down : decrease simulation speed
        - space : pause simulation
        """
        #if symbol & key.BACKSPACE or (symbol & key._0 and (modifiers & MOD_COMMAND or modifiers & MOD_CTRL)):
        if symbol == key.BACKSPACE or (symbol == key._0 and (modifiers & key.MOD_COMMAND)):
            self.left = self.orig_left
            self.right = self.orig_right
            self.bottom = self.orig_bottom
            self.top = self.orig_top
            self.zoom_level = self.orig_zoom_level
            self.zoomed_width = self.orig_zoomed_width
            self.zoomed_height = self.orig_zoomed_height

            glMatrixMode( GL_PROJECTION )
            glLoadIdentity()
            glOrtho( self.left, self.right, self.bottom, self.top, 1, -1 )
        elif symbol == key.UP:
            self.simulation_status.sampling_dt *= 1.2
        elif symbol == key.DOWN:
            self.simulation_status.sampling_dt /= 1.2
        elif symbol == key.SPACE:
            self.simulation_status.paused = not self.simulation_status.paused



class Scale():
    """
    A scale that maps all its connected graphics objects
    to world (window) dimensions.

    Parameters
    ==========
    bound_increase_factor : float, default = 1.0
        By how much the respective bound should increase
        once it's reached.

    Attributes
    ==========
    bound_increase_factor : float
        By how much the respective bound should increase
        once it's reached.
    x0 : float
        lower bound of data x-dimension
    x1 : float
        upper bound of data x-dimension
    y0 : float
        lower bound of data y-dimension
    y1 : float
        upper bound of data y-dimension
    left : float
        lower bound of world x-dimension
    right : float
        upper bound of world x-dimension
    bottom : float
        lower bound of world y-dimension
    top : float
        upper bound of world y-dimension
    scaling_objects : list
        A list of objects that need to be rescaled
        once the data or world dimensions change.
        Each entry of this list is assumed to be 
        an object that has a method called ``rescale()``.
    """

    def __init__(self,bound_increase_factor=1.0):

        self.x0 = np.nan
        self.y0 = np.nan
        self.x1 = np.nan
        self.y1 = np.nan
        self.left = np.nan
        self.right = np.nan
        self.bottom = np.nan
        self.top = np.nan

        self.bound_increase_factor = bound_increase_factor

        self.scaling_objects = []

    def extent(self,left,right,top,bottom):
        """
        Define the world (window) dimensions.
        """
        self.left = left
        self.right = right
        self.top = top
        self.bottom = bottom
        self._calc()
        return self

    def domain(self,x0,x1,y0,y1):
        """
        Define the data dimensions.
        """
        self.x0 = x0
        self.x1 = x1
        self.y0 = y0
        self.y1 = y1
        self._calc()
        return self

    def _calc(self):
        """
        calculate scalars
        """
        self.mx = (self.right-self.left)/(self.x1 - self.x0)
        self.my = (self.top-self.bottom)/(self.y1 - self.y0)

    def scale(self,x,y):
        """
        Scale data.
        """
        _x = self.scalex(x)
        _y = self.scaley(y)

        return _x, _y

    def scalex(self,x):
        """
        Scale x-data
        """
        if type(x) == list:
            _x = list(map(lambda _x: self.mx * (_x-self.x0) + self.left, x ))
        else:
            _x = self.mx * (x-self.x0) + self.left

        return _x

    def scaley(self,y):
        """
        Scale y-data
        """
        if type(y) == list:
            _y = list(map(lambda _y: self.my * (_y-self.y0) + self.bottom, y ))
        else:
            _y = self.my * (y-self.y0) + self.bottom

        return _y

    def check_bounds(self,xmin,xmax,ymin,ymax):
        """
        Check whether the global data dimensions have changed
        considered updated data dimensions of a single instance.
        If this is the case, trigger rescaling of all connected
        instances.
        """
        changed = False
        x0, x1, y0, y1 = self.x0, self.x1, self.y0, self.y1
        if xmin < self.x0:
            xvec = xmin - self.x1
            x0 = self.x1 + xvec * self.bound_increase_factor
            changed = True
        if ymin < self.y0:
            yvec = ymin - self.y1
            y0 = self.y1 + yvec * self.bound_increase_factor
            changed = True
        if xmax > self.x1:
            xvec = xmax - self.x0
            x1 = self.x0 + xvec * self.bound_increase_factor
            changed = True
        if ymax > self.y1:
            yvec = ymax - self.y0
            y1 = self.y0 + yvec * self.bound_increase_factor
            changed = True

        if changed:
            self.domain(x0,x1,y0,y1)
            for obj in self.scaling_objects:
                obj.rescale()

    def add_scaling_object(self,obj):
        """
        Append an object that depends on this Scale instance.
        """
        self.scaling_objects.append(obj)



class Curve():
    """
    A class that draws an OpenGL
    curve to a pyglet Batch instance
    with easy methods to update data.

    Parameters
    ==========
        x : list
            x data
        y : list
            y data
        color : list 
            List of 3 integers between 0 and 255 (RGB color list)
        scale : Scale
            An instance of a Scale that maps the data dimensions
            to an area in a pyglet Window

    Attributes
    ==========
        batch : pyglet.graphics.Batch
            The batch instance in which this curve is drawn
        scale : Scale
            An instance of a Scale that maps the data dimensions
            to an area in a pyglet Window
        vertex_list : pyglet.graphics.VertexList
            Contains the vertex list in window coordinates.
            format strings are ``v2f`` and ``c3B``.
        color : list
            as described above
        xmin : float
            lower bound of x-dimension
        xmax : float
            upper bound of x-dimension
        ymin : float
            lower bound of y-dimension
        ymax : float
            upper bound of y-dimension
    """

    def __init__(self,x,y,color,scale,batch):
        self.batch = batch
        self.vertex_list = batch.add(0, GL_LINE_STRIP, None,
                    'v2f',
                    'c3B',
                )
        self.scale = scale
        scale.add_scaling_object(self)
        self.color = color
        self.xmin = 1e300
        self.ymin = 1e300
        self.xmax = -1e300
        self.ymax = -1e300
        self.set(x,y)

    def set(self,x,y):
        """
        Set the data of this curve.
        """
        _x = list(x)
        _y = list(y)

        # GL_LINE_STRIP needs to have the first
        # and last vertex as duplicates,
        # save data
        self.x = _x[:1] + _x + _x[-1:]
        self.y = _y[:1] + _y + _y[-1:]

        # get min/max values of this update
        xmin = min(_x)
        xmax = max(_x)
        ymin = min(_y)
        ymax = max(_y)

        # scale and zip together the new numbers
        _x, _y = self.scale.scale(self.x, self.y)
        xy = list(chain.from_iterable(zip(_x, _y)))

        # resize vertex list, set vertices and colors
        self.vertex_list.resize(len(_x))
        self.vertex_list.vertices = xy
        self.vertex_list.colors = self.color * len(_x)

        # check whether or not the bounds of
        # the scale need to be updated
        self.update_bounds(xmin,xmax,ymin,ymax)

    def append_single_value(self,x,y):
        """
        Append a single data point to this curve.
        Note that if the bounds change with this
        update, the connect Scale-instance will be updated
        automatically.
        """
        self.append_list([x], [y])

    def append_list(self,x,y):
        """
        Append a list of data points to this curve.
        Note that if the bounds change with this
        update, the connect Scale-instance will be updated
        automatically.
        """
        _x = x + x[-1:]
        _y = y + y[-1:]

        # remember that self.x contains the last
        # vertex twice for GL_LINE_STRIP.
        # We have to pop the duplicate of the
        # formerly last entry and append the new list
        self.x.pop()
        self.x.extend(_x)
        self.y.pop()
        self.y.extend(_y)

        xmin = min(_x)
        xmax = max(_x)
        ymin = min(_y)
        ymax = max(_y)

        _x, _y = self.scale.scale(_x, _y)
        xy = list(chain.from_iterable(zip(_x, _y)))

        self.vertex_list.resize(self.vertex_list.get_size() + len(_x) -1 )
        self.vertex_list.vertices[-len(xy):] = xy
        self.vertex_list.colors[-3*len(_x):] = self.color * len(_x)

        self.update_bounds(xmin,xmax,ymin,ymax)

    def rescale(self):
        """
        Rescale this curve's data according to
        the connected Scale-instance
        """
        _x, _y = self.scale.scale(self.x, self.y)
        xy = list(chain.from_iterable(zip(_x, _y)))
        self.vertex_list.vertices = xy

    def update_bounds(self,xmin,xmax,ymin,ymax):
        """
        Compute the bounds of this curves data and
        update the scale accordingly
        """
        self.xmin = min(self.xmin,xmin)
        self.ymin = min(self.ymin,ymin)
        self.xmax = max(self.xmax,xmax)
        self.ymax = max(self.ymax,ymax)

        self.scale.check_bounds(self.xmin,self.xmax,self.ymin,self.ymax)

def get_network_batch(stylized_network,
                      yoffset,
                      draw_links=True,
                      draw_nodes=True,
                      draw_nodes_as_rectangles=False,
                      n_circle_segments=16):
    """
    Create a batch for a network visualization.

    Parameters
    ----------
    stylized_network : dict
        The network properties which are returned from the
        interactive visualization.
    draw_links : bool, default : True
        Whether the links should be drawn
    draw_nodes : bool, default : True
        Whether the nodes should be drawn
    n_circle_segments : bool, default = 16
        Number of segments a circle will be constructed of.

    Returns
    -------
    network_objects : dict
        A dictionary containing all the necessary objects to draw and
        update the network.
        - `lines` : a list of pyglet-line objects (one entry per link)

        - `disks` : a list of pyglet-circle objects (one entry per node)
        - `circles` : a list of pyglet-circle objects (one entry per node)
        - `nodes_to_lines` : a dictionary mapping a node to a list of
          pairs. Each pair's first entry is the focal node's neighbor
          and the second entry is the index of the line-object that
          connects the two
        - `batch` : the pyglet Batch instance that contains all of the objects
    """

    batch = pyglet.graphics.Batch()

    pos = { node['id']: np.array([node['x_canvas'], node['y_canvas'] + yoffset]) for node in stylized_network['nodes'] }

    lines = []
    disks = []
    circles = []
    node_to_lines = { node['id']: [] for node in stylized_network['nodes'] }

    if draw_links:

        for ilink, link in enumerate(stylized_network['links']):
            u, v = link['source'], link['target']
            node_to_lines[u].append((v, ilink))
            node_to_lines[v].append((u, ilink))
            if 'color' in link.keys():
                this_color = link['color']
            else:
                this_color = stylized_network['linkColor']
            lines.append(shapes.Line(
                pos[u][0],
                pos[u][1],
                pos[v][0],
                pos[v][1],
                width=link['width'],
                color=tuple(bytes.fromhex(this_color[1:])),
                batch=batch,
                         )
                    )
            lines[-1].opacity = int(255*stylized_network['linkAlpha'])


    if draw_nodes:
        disks = [None for n in range(len(stylized_network['nodes']))]
        circles = [None for n in range(len(stylized_network['nodes']))]


        for node in stylized_network['nodes']:
            if not draw_nodes_as_rectangles:
                disks[node['id']] = \
                        shapes.Circle(node['x_canvas'],
                                      node['y_canvas']+yoffset,
                                      node['radius'],
                                      segments=n_circle_segments,
                                      color=tuple(bytes.fromhex(node['color'][1:])),
                                      batch=batch,
                                      )

                circles[node['id']] = \
                        shapes.Arc(node['x_canvas'],
                                      node['y_canvas']+yoffset,
                                      node['radius'],
                                      segments=n_circle_segments+1,
                                      color=tuple(bytes.fromhex(stylized_network['nodeStrokeColor'][1:])),
                                      batch=batch,
                                      )
            else:
                r = node['radius']
                disks[node['id']] = \
                        shapes.Rectangle(
                                      node['x_canvas']-r,
                                      node['y_canvas']+yoffset-r,
                                      2*r,
                                      2*r,
                                      color=tuple(bytes.fromhex(node['color'][1:])),
                                      batch=batch)

    return {'lines': lines, 'disks': disks, 'circles':circles, 'node_to_lines': node_to_lines, 'batch': batch}

_default_config = {
            'plot_sampled_curve': True,
            'draw_links':True,
            'draw_nodes':True,
            'n_circle_segments':16,
            'plot_height':120,
            'bgcolor':'#253237',
            'curve_stroke_width':4.0,
            'node_stroke_width':1.0,
            'link_color': '#4b5a62',
            'node_stroke_color':'#000000',
            'node_color':'#264653',
            'bound_increase_factor':1.0,
            'update_dt':0.04,
            'show_curves':True,
            'draw_nodes_as_rectangles':False,
            'show_legend': True,
            'legend_font_color':None,
            'legend_font_size':10,
            'padding':10,
            'compartment_colors':_colors,
            'palette': "dark",
        }

# light colors
#_default_config.update({
#            'bgcolor':'#fbfbef',
#            'link_color': '#8e9aaf',
#            'node_stroke_color':'#000000',
#            'legend_font_color':'#040414',
#        })


def visualize(model,
              network, 
              sampling_dt,
              ignore_plot_compartments=[],
              quarantine_compartments=[],
              config=None,
              ):
    """
    Start a visualization of a stochastic simulation.

    Parameters
    ==========
    model : epipack.stochastic_epi_models.StochasticEpiModel
        An initialized StochasticEpiModel.
    network: dict
        A stylized network in the netwulf-format
        (see https://netwulf.readthedocs.io/en/latest/python_api/post_back.html)
        where instead of 'x' and 'y', node positions are saved in 'x_canvas'
        and 'y_canvas'. Example:

        .. code:: python

            stylized_network = {
                "xlim": [0, 833],
                "ylim": [0, 833],
                "linkAlpha": 0.5,
                "nodeStrokeWidth": 0.75,
                "links": [
                    {"source": 0, "target": 1, "width": 3.0 }
                ],
                "nodes": [
                    {"id": 0,
                     "x_canvas": 436.0933431058901,
                     "y_canvas": 431.72418500564186,
                     "radius": 20
                     },
                    {"id": 1,
                     "x_canvas": 404.62184898400426,
                     "y_canvas": 394.8158724310507,
                     "radius": 20
                     }
                ]
            }

    sampling_dt : float 
        The amount of simulation time that's supposed to pass
        with a single update.
    ignore_plot_compartments : list, default = []
        List of compartment objects that are supposed to be
        ignored when plotted.
    quarantine_compartments : list, default = []
        List of compartment objects that are supposed to be
        resemble quarantine (i.e. temporarily 
        losing all connections)
    config : dict, default = None
        A dictionary containing all possible configuration
        options. Entries in this dictionary will overwrite
        the default config which is 

        .. code:: python

            _default_config = {
                        'plot_sampled_curve': True,
                        'draw_links':True,
                        'draw_nodes':True,
                        'n_circle_segments':16,
                        'plot_height':120,
                        'bgcolor':'#253237',
                        'curve_stroke_width':4.0,
                        'node_stroke_width':1.0,
                        'link_color': '#4b5a62',
                        'node_stroke_color':'#000000',
                        'node_color':'#264653',
                        'bound_increase_factor':1.0,
                        'update_dt':0.04,
                        'show_curves':True,
                        'draw_nodes_as_rectangles':False,
                        'show_legend': True,
                        'legend_font_color':'#fafaef',
                        'legend_font_size':10,
                        'padding':10,
                        'compartment_colors':_colors
                    }

    """

    # update the config and compute some helper variables
    cfg = deepcopy(_default_config)
    if config is not None:
        cfg.update(config)

    palette = cfg['palette']
    if type(palette) == str:
        if 'link_color' not in cfg:
            cfg['link_color'] = col.hex_link_colors[palette]
        if 'bgcolor' not in cfg:
            cfg['bgcolor'] = col.hex_bg_colors[palette]
        if 'compartment_colors' not in cfg:
            cfg['compartment_colors'] = [ col.colors[this_color] for this_color in col.palettes[palette] ]

    bgcolor = [ _/255 for _ in list(bytes.fromhex(cfg['bgcolor'][1:])) ] + [1.0]

    bgY = 0.2126*bgcolor[0] + 0.7152*bgcolor[1] + 0.0722*bgcolor[2]
    if cfg['legend_font_color'] is None:
        if bgY < 0.5:
            cfg['legend_font_color'] = '#fafaef'
        else:
            cfg['legend_font_color'] = '#232323'

    width = network['xlim'][1] - network['xlim'][0]
    height = network['ylim'][1] - network['ylim'][0]

    with_plot = cfg['show_curves'] and set(ignore_plot_compartments) != set(model.compartments)

    if with_plot:
        height += cfg['plot_height']
        plot_width = width
        plot_height = cfg['plot_height']
    else:
        plot_height = 0

    with_legend = cfg['show_legend']

    if with_legend:
        legend_batch = pyglet.graphics.Batch()
        #x, y = legend.get_location()
        #legend.set_location(x - width, y)
        # create a test label to get the actual dimensions
        test_label = pyglet.text.Label('Ag')
        dy = test_label.content_height * 1.1
        del(test_label)

        legend_circle_radius = dy/2/2
        distance_between_circle_and_label = 2*legend_circle_radius
        legend_height = len(model.compartments) * dy + cfg['padding']

        # if legend is shown in concurrence to the plot,
        # move the legend to be on the right hand side of the plot,
        # accordingly make the plot at least as tall as 
        # the demanded height or the legend height
        if with_plot:
            plot_height = max(plot_height, legend_height)
        legend_y_offset = legend_height

        max_text_width = 0
        legend_objects = [] # this is a hack so that the garbage collector doesn't delete our stuff 
        for iC, C in enumerate(model.compartments):
            this_y = legend_y_offset - iC * dy - cfg['padding']
            this_x = width + cfg['padding'] + legend_circle_radius
            label = pyglet.text.Label(str(C),
                              font_name=('Helvetica', 'Arial', 'Sans'),
                              font_size=cfg['legend_font_size'],
                              x=this_x + legend_circle_radius+distance_between_circle_and_label, 
                              y=this_y,
                              anchor_x='left', anchor_y='top',
                              color = list(bytes.fromhex(cfg['legend_font_color'][1:])) + [255],
                              batch = legend_batch
                              )
            legend_objects.append(label)

            #if not cfg['draw_nodes_as_rectangles']:
            if True:
                disk = shapes.Circle(this_x,
                                      this_y - (dy-1.25*legend_circle_radius)/2,
                                      legend_circle_radius,
                                      segments=64,
                                      color = cfg['compartment_colors'][iC],
                                      batch=legend_batch,
                                      )

                circle = shapes.Arc(this_x,
                                      this_y - (dy-1.25*legend_circle_radius)/2,
                                      legend_circle_radius,
                                      segments=64+1,
                                      color=list(bytes.fromhex(cfg['legend_font_color'][1:])),
                                      batch=legend_batch,
                                      )

                legend_objects.extend([disk,circle])
            #else:
            #    rect = shapes.Rectangle(this_x,
            #                          this_y - (dy-1.5*legend_circle_radius)/2,
            #                          2*legend_circle_radius,
            #                          2*legend_circle_radius,
            #                          color = _colors[iC],
            #                          batch=legend_batch,
            #                          )
            #    legend_objects.append(rect)

            max_text_width = max(max_text_width, label.content_width)

        legend_width =   2*cfg['padding'] \
                       + 2*legend_circle_radius \
                       + distance_between_circle_and_label \
                       + max_text_width

        # if legend is shown in concurrence to the plot,
        # move the legend to be on the right hand side of the plot,
        # accordingly make the plot narrower and place the legend
        # directly under the square network plot.
        # if not, make the window wider and show the legend on
        # the right hand side of the network plot.
        if with_plot:
            for obj in legend_objects:
                obj.x -= legend_width 
            plot_width = width - legend_width
        else:
            width += legend_width

    size = (width, height)


    # overwrite network style with the epipack default style
    network['linkColor'] = cfg['link_color']
    network['nodeStrokeColor'] = cfg['node_stroke_color']
    for node in network['nodes']:
        node['color'] = cfg['node_color']
    N = len(network['nodes'])

    # get the OpenGL shape objects that comprise the network
    network_batch = get_network_batch(network,
                                      yoffset=plot_height,
                                      draw_links=cfg['draw_links'],
                                      draw_nodes=cfg['draw_nodes'],
                                      draw_nodes_as_rectangles=cfg['draw_nodes_as_rectangles'],
                                      n_circle_segments=cfg['n_circle_segments'],
                                      )
    lines = network_batch['lines']
    disks = network_batch['disks']
    circles = network_batch['circles']
    node_to_lines = network_batch['node_to_lines']
    batch = network_batch['batch']

    # initialize a simulation state that has to passed to the app
    # so the app can change simulation parameters
    simstate = SimulationStatus(len(network['nodes']), sampling_dt)

    # intialize app
    window = App(*size,simulation_status=simstate,resizable=True)
    glClearColor(*bgcolor)

    # handle different strokewidths
    if 'nodeStrokeWidth' in network:
        node_stroke_width = network['nodeStrokeWidth'] 
    else:
        node_stroke_width = cfg['node_stroke_width']

    def _set_linewidth_nodes():
        glLineWidth(node_stroke_width)

    def _set_linewidth_curves():
        glLineWidth(cfg['curve_stroke_width'])

    def _set_linewidth_legend():
        glLineWidth(1.0)

    # add the network batch with the right function to set the linewidth
    # prior to drawing
    window.add_batch(batch,prefunc=_set_linewidth_nodes)

    if with_legend:
        # add the legend batch with the right function to set the linewidth
        # prior to drawing
        window.add_batch(legend_batch,prefunc=_set_linewidth_legend)

    # decide whether to plot all measured changes or only discrete-time samples
    discrete_plot = cfg['plot_sampled_curve']

    # find quarantined compartment ids
    # This set is needed for filtering later on.
    quarantined = set(model.get_compartment_id(C) for C in quarantine_compartments)

    # initialize time arrays
    t = 0
    discrete_time = [t]

    # initialize curves
    if with_plot:
        # find the maximal value of the
        # compartments that are meant to be plotted. 
        # These sets are needed for filtering later on.
        maxy = max([ model.y0[model.get_compartment_id(C) ] for C in (set(model.compartments) - set(ignore_plot_compartments))])
        scl = Scale(bound_increase_factor=cfg['bound_increase_factor'])\
                .extent(0,plot_width,plot_height-cfg['padding'],cfg['padding'])\
                .domain(0,20*sampling_dt,0,maxy)
        curves = {}
        for iC, C in enumerate(model.compartments):
            if C in ignore_plot_compartments:
                continue
            _batch = pyglet.graphics.Batch()
            window.add_batch(_batch,prefunc=_set_linewidth_curves)
            y = [np.count_nonzero(model.node_status==model.get_compartment_id(C))]
            curve = Curve(discrete_time,y,cfg['compartment_colors'][iC],scl,_batch)
            curves[C] = curve

    # define the pyglet-App update function that's called on every clock cycle
    def update(dt):

        # skip if nothing remains to be done
        if simstate.simulation_ended or simstate.paused:
            return

        # get sampling_dt
        sampling_dt = simstate.sampling_dt

        # Advance the simulation until time sampling_dt.
        # sim_time is a numpy array including all time values at which
        # the system state changed. The first entry is the initial state
        # of the simulation at t = 0 which we will discard later on
        # the last entry at `sampling_dt` will be missing so we
        # have to add it later on.
        # `sim_result` is a dictionary that maps a compartment
        # to a numpy array containing the compartment counts at
        # the corresponding time.
        sim_time, sim_result = model.simulate(sampling_dt,adopt_final_state=True)

        # compare the new node statuses with the old node statuses
        # and find the nodes that have changed status
        ndx = np.where(model.node_status != simstate.old_node_status)[0]

        # if nothing changed, evaluate the true total event rate
        # and if it's zero, do not do anything anymore
        did_simulation_end = len(ndx) == 0 and model.get_true_total_event_rate() == 0.0
        simstate.set_simulation_status(did_simulation_end)
        if simstate.simulation_ended:
            return

        # advance the current time as described above.
        # we save both all time values as well as just the sampled times.
        this_time = (discrete_time[-1] + sim_time[1:]).tolist() + [discrete_time[-1] + sampling_dt]
        discrete_time.append(discrete_time[-1] + sampling_dt)

        # if curves are plotted
        if with_plot:

            # iterate through result array
            for k, v in sim_result.items():
                # skip curves that should be ignored
                if k in ignore_plot_compartments:
                    continue
                # count occurrences of this compartment
                val = np.count_nonzero(model.node_status==model.get_compartment_id(k))
                if discrete_plot:
                    # in case only sampled curves are of interest,
                    # just add this single value
                    curves[k].append_single_value(discrete_time[-1], v[-1])
                else:
                    # otherwise, append the current value to the exact simulation list
                    # and append the whole dataset
                    val = (v[1:].tolist() + [v[-1]])
                    curves[k].append_list(this_time, val)


        # iterate through the nodes that have to be updated
        for node in ndx:
            status = model.node_status[node]
            if cfg['draw_nodes']:
                disks[node].color = cfg['compartment_colors'][status]

            # if a node becomes quarantined,
            # iterate through its attached links (lines)
            # and switch them off
            if status in quarantined:
                for neigh, linkid in node_to_lines[node]:
                    lines[linkid].visible = False
            # if it became unquarantined
            elif simstate.old_node_status[node] in quarantined:
                # check of the neighbor is unquarantined
                # and switch on the link if this is the case
                for neigh, linkid in node_to_lines[node]:
                    if model.node_status[neigh] not in quarantined:
                        lines[linkid].visible = True

        # save the current node statuses
        simstate.update(model.node_status)

    # schedule the app clock and run the app
    pyglet.clock.schedule_interval(update, cfg['update_dt'])
    pyglet.app.run()
    #pyglet.clock.unschedule(update)
    #window.close()
    #del(window)
    #print(pyglet.app.platform_event_loop)
    #pyglet.app.exit()
    #print("called exit..")


def visualize_reaction_diffusion(
              model,
              network,
              sampling_dt,
              node_compartments,
              value_extent=[0.0,1.0],
              integrator='euler',
              n_integrator_midpoints=0,
              config=None,
              ):
    """
    Start a visualization of a reaction-diffusion simulation.

    Parameters
    ==========
    model : epipack.deterministic_epi_models.DeterministicEpiModel
        An initialized DeterministicEpiModel.
    network: dict
        A stylized network in the netwulf-format
        (see https://netwulf.readthedocs.io/en/latest/python_api/post_back.html)
        where instead of 'x' and 'y', node positions are saved in 'x_canvas'
        and 'y_canvas'. Example:

        .. code:: python

            stylized_network = {
                "xlim": [0, 833],
                "ylim": [0, 833],
                "linkAlpha": 0.5,
                "nodeStrokeWidth": 0.75,
                "links": [
                    {"source": 0, "target": 1, "width": 3.0 }
                ],
                "nodes": [
                    {"id": 0,
                     "x_canvas": 436.0933431058901,
                     "y_canvas": 431.72418500564186,
                     "radius": 20
                     },
                    {"id": 1,
                     "x_canvas": 404.62184898400426,
                     "y_canvas": 394.8158724310507,
                     "radius": 20
                     }
                ]
            }

    sampling_dt : float
        The amount of simulation time that's supposed to pass
        with a single update.
    quarantine_compartments: list
        List of compartment objects that are supposed to be
        resemble quarantine (i.e. temporarily
        losing all connections)
    node_compartments: list
        The compartments for which to display the concentrations.
        Each entry `m` in this list is expected to be a compartment
        associated with node `m`. this list should therefore
        be as long as the number of nodes.
    config : dict, default = None
        A dictionary containing all possible configuration
        options. Entries in this dictionary will overwrite
        the default config which is

        .. code:: python

            _default_config = {
                        'plot_sampled_curve': True,
                        'draw_links':True,
                        'draw_nodes':True,
                        'n_circle_segments':16,
                        'plot_height':120,
                        'bgcolor':'#253237',
                        'curve_stroke_width':4.0,
                        'node_stroke_width':1.0,
                        'link_color': '#4b5a62',
                        'node_stroke_color':'#000000',
                        'node_color':'#264653',
                        'bound_increase_factor':1.0,
                        'update_dt':0.04,
                        'show_curves':True,
                        'draw_nodes_as_rectangles':False,
                        'show_legend': True,
                        'legend_font_color':'#fafaef',
                        'legend_font_size':10,
                        'padding':10,
                        'compartment_colors':_colors
                    }

    """

    # update the config and compute some helper variables
    cfg = deepcopy(_default_config)
    if config is not None:
        cfg.update(config)

    bgcolor = [ _/255 for _ in list(bytes.fromhex(cfg['bgcolor'][1:])) ] + [1.0]

    width = network['xlim'][1] - network['xlim'][0]
    height = network['ylim'][1] - network['ylim'][0]

    size = (width, height)



    # overwrite network style with the epipack default style
    network['linkColor'] = cfg['link_color']
    network['nodeStrokeColor'] = cfg['node_stroke_color']
    for node in network['nodes']:
        node['color'] = col.hex_bg_colors['light']
    N = len(network['nodes'])

    # get the OpenGL shape objects that comprise the network
    network_batch = get_network_batch(network,
                                      yoffset=0,
                                      draw_links=cfg['draw_links'],
                                      draw_nodes=cfg['draw_nodes'],
                                      draw_nodes_as_rectangles=cfg['draw_nodes_as_rectangles'],
                                      n_circle_segments=cfg['n_circle_segments'],
                                      )
    lines = network_batch['lines']
    disks = network_batch['disks']
    circles = network_batch['circles']
    node_to_lines = network_batch['node_to_lines']
    batch = network_batch['batch']

    # initialize a simulation state that has to passed to the app
    # so the app can change simulation parameters
    simstate = SimulationStatus(len(network['nodes']), sampling_dt)

    # intialize app
    window = App(*size,simulation_status=simstate,resizable=True)
    glClearColor(*bgcolor)

    # handle different strokewidths
    if 'nodeStrokeWidth' in network:
        node_stroke_width = network['nodeStrokeWidth']
    else:
        node_stroke_width = cfg['node_stroke_width']

    def _set_linewidth_nodes():
        glLineWidth(node_stroke_width)

    # add the network batch with the right function to set the linewidth
    # prior to drawing
    window.add_batch(batch,prefunc=_set_linewidth_nodes)

    # initialize time arrays
    t = 0
    discrete_time = [t]

    _cmin, _cmax = value_extent
    if not callable(_cmin):
        cmin = lambda _: _cmin
    else:
        cmin = _cmin
    if not callable(_cmax):
        cmax = lambda _: _cmax
    else:
        cmax = _cmax
    def _get_opacity(val,this_cmin,this_cmax):
        opacity = (val-this_cmin)/(this_cmax-this_cmin) + this_cmin
        if opacity > 1.0:
            opacity = 1.0
        if opacity < 0.0:
            opacity = 0.0
        return int(255*opacity)

    try:
        node_compartment_indices = np.array([ model.get_compartment_id(C) for C in node_compartments ])
    except AttributeError as e:
        node_compartment_indices = np.array(node_compartments)

    this_cmin, this_cmax = cmin(model.y0), cmax(model.y0)
    for node, idx in enumerate(node_compartment_indices):
        concentration = model.y0[idx]
        disks[node].opacity = _get_opacity(concentration, this_cmin, this_cmax)

    # define the pyglet-App update function that's called on every clock cycle
    def update(dt):

        # skip if nothing remains to be done
        if simstate.simulation_ended or simstate.paused:
            return

        # get sampling_dt
        sampling_dt = simstate.sampling_dt

        # Advance the simulation until time sampling_dt.
        # `sim_result` is a two-dimensional array
        # where index sim_result[iC, iT] gives the
        # concentration of compartment iC at time iT
        this_t = np.linspace(0, sampling_dt, n_integrator_midpoints+2)
        sim_result = model.integrate_and_return_by_index(
                            this_t,
                            integrator=integrator,
                            adopt_final_state=True,
                            )

        # it might happen that an external model only returns
        # the final state instead of for each time point
        if len(sim_result.shape) > 1:
            result = sim_result[node_compartment_indices,-1]
        else:
            result = sim_result[node_compartment_indices]

        # if nothing changed, evaluate the true total event rate
        # and if it's zero, do not do anything anymore
        #did_simulation_end = len(ndx) == 0 and model.get_true_total_event_rate() == 0.0
        #simstate.set_simulation_status(did_simulation_end)
        #if simstate.simulation_ended:
        #    return
        this_cmin, this_cmax = cmin(result), cmax(result)

        # detect significant changes (relative change)
        rel_change = np.abs(result-simstate.old_node_status)
        ndx = np.where(rel_change>1e-3)[0]
        #ndx = np.where(rel_change>0)[0]
        result = result[ndx]

        # iterate through the nodes that have to be updated
        #for node, idx in enumerate(node_compartment_indices):
        if cfg['draw_nodes']:
            for node, concentration in zip(ndx, result):
                    disks[node].opacity = _get_opacity(concentration,this_cmin,this_cmax)

        simstate.old_node_status[ndx] = result
        #simstate.update(result)


    # schedule the app clock and run the app
    pyglet.clock.schedule_interval(update, cfg['update_dt'])
    pyglet.app.run()


if __name__=="__main__":     # pragma: no cover

    import netwulf as nw
    from epipack import StochasticEpiModel

    network, config, _ = nw.load('/Users/bfmaier/pythonlib/facebook/FB.json')

    N = len(network['nodes'])

    edge_list = [ ( link['source'], link['target'], 1.0 ) for link in network['links'] ]
    k0 = 2*len(edge_list)/len(network['nodes'])

    model = StochasticEpiModel(list("SIRXTQ"),
                               N=len(network['nodes']),
                               edge_weight_tuples=edge_list,
                               )
    Reff = 3
    R0 = 10
    recovery_rate = 1/8
    quarantine_rate = 1/16
    tracing_rate = 1/2
    waning_immunity_rate = 1/14
    infection_rate = Reff * (recovery_rate+quarantine_rate) / k0
    infection_rate = R0 * (recovery_rate) / k0
    model.set_node_transition_processes([
            ("I",recovery_rate,"R"),
            ("I",quarantine_rate,"T"),
            ("T",tracing_rate,"X"),
            ("Q",waning_immunity_rate,"S"),
            ("X",recovery_rate,"R"),
            ])
    model.set_link_transmission_processes([("I","S",infection_rate,"I","I")])
    model.set_conditional_link_transmission_processes({
        ("T", "->", "X") : [
                 ("X","I",0.5,"X","T"),
                 #("X","S",0.5,"X","Q"),
                 ],
        })
    model.set_random_initial_conditions({'I':20,'S':N-20})

    sampling_dt = 0.08

    visualize(model,network,sampling_dt,ignore_plot_compartments=['S'],quarantine_compartments=['X', 'T', 'Q'])




"""
Classes to define temporal networks and run stochastic simulations
on them.
"""

import numpy as np

class TemporalNetwork():
    """
    A simple temporal network class.

    Parameters
    ==========
    N : int
        Static number of nodes in the network.
    t : list
        An increasingly ordered list of times.
    tmax : float
        The time at which the recording of network changes
        has concluded (the final time of the
        temporal network).
    edge_lists : list
        A list of lists of tuples. Each entry i
        of this list represents an edge list.
        Each entry j of an edge list is a tuple
        of format

        .. code:: python

            ( source, target, weight )

        if ``weighted=True``, or

        .. code:: python

            ( source, target )

        otherwise. ``source`` and ``target``
        are supposed to be of unsigned integer type
        in range ``[0,N)``.
    weighted : bool, default = False
        Whether or not ``edge_lists`` contain edge
        tuples with weights
    directed : bool, default = False
        Whether or not edge tuples will be considered
        to be directed.
    loop_network : bool, default = True
        If `True`, the temporal network will be looped
        indefinitely when iterated.

    Example
    =======

    >>> edges = [ [ (0,1) ], [ (0,1), (0,2) ], [] ]
    >>> temporal_network = TemporalNetwork(3,[0,0.5,0.6],1.0,edges)
    >>> for edge_list, t, next_t in temporal_network:
    ...     if t >= 3:
    ...         break
    ...     print(t, next_t, edge_list)
    ...
    0 0.5 [(0, 1, 1.0)]
    0.5 0.6 [(0, 1, 1.0), (0, 2, 1.0)]
    0.6 1.0 []
    1.0 1.5 [(0, 1, 1.0)]
    1.5 1.6 [(0, 1, 1.0), (0, 2, 1.0)]
    1.6 2.0 []
    2.0 2.5 [(0, 1, 1.0)]
    2.5 2.6 [(0, 1, 1.0), (0, 2, 1.0)]
    2.6 3.0 []
    

    Attributes
    ==========
    N : int
        Total and constant number of nodes in the system.
    t : list
        An increasingly ordered list of times.
    tmax : float
        The time at which recording of edge changes
        has concluded (the final time of the
        temporal network).
    edge_lists : list
        A list of lists of tuples. Each entry i
        of this list represents an edge list.
        Each entry j of an edge list is a tuple
        of format

        .. code:: python

            ( source, target, weight )

        if ``weighted=True``, or

        .. code:: python

            ( source, target )

        otherwise. ``source`` and ``target``
        are supposed to be of unsigned integer type
        in range ``[0,N)``.
    weighted : bool
        Whether or not ``edge_lists`` contain edge
        tuples with weights
    directed : bool
        Whether or not edge uples will be considered
        to be directed.
    loop_network : bool
        If `True`, the temporal network will be looped
        indefinitely when iterated.
    """

    def __init__(self,
                 N,
                 edge_lists,
                 t,
                 tmax,
                 weighted=False,
                 directed=False,
                 loop_network=True,
                 ):

        assert([ a == b for a, b in zip(sorted(t),t) ])
        assert(t[-1] < tmax)
        assert(len(edge_lists) == len(t))

        _t = np.array(t)
        assert(np.all(_t[1:] - _t[:-1] > 0))

        self.t = t
        self.tmax = tmax
        self.N = N
        self.edge_lists = edge_lists
        self.weighted = weighted
        self.directed = directed
        self.T = tmax - t[0]
        self.loop_network = loop_network

    @classmethod
    def from_tacoma(cls, edge_lists, loop_network=True):
        """Initiate from a tacoma.edge_lists instance"""
        return cls(edge_lists.N,
                   edge_lists.edges,
                   edge_lists.t,
                   edge_lists.tmax,
                   weighted=False,
                   directed=False,
                   loop_network=loop_network,
                   )

    def t0(self):
        """Obtain the initial time"""
        return self.t[0]

    def mean_out_degree(self):
        """Obtain the temporally averaged mean out-degree"""
        if not self.weighted:
            k = 0.0
            if not self.directed:
                factor = 2
            else:
                factor = 1
            for edges, t, next_t in self:
                if self._has_looped():
                    break
                k += (next_t - t) * factor*len(edges)/self.N
            k /= self.T
        else:
            k = np.zeros(self.N,dtype=float)
            for edges, t, next_t in self:
                if self._has_looped():
                    break
                dt = next_t - t
                for u, v, w in edges:
                    k[u] += w * dt
                    if not self.directed:
                        k[v] += w * dt
            k = k.mean() / self.T
        return k

    def _has_looped(self):
        """Whether or not the network has looped at least once"""
        return self._Delta_t >= self.T

    def __iter__(self):
        self._ndx = 0
        self._Delta_t = 0
        return self

    def __next__(self):

        if self._ndx < len(self.t):

            t0 = self.t0()

            # get the current time (including potential loops)
            t = self.t[self._ndx] - t0
            t += self._Delta_t + t0

            # get the current edges and add default weight = 1.0 if unweighted
            edges = self.edge_lists[self._ndx]
            if not self.weighted:
                edges = [ (u,v,1.0) for u, v in edges ]

            # increase the iteration pointer and calculate the next
            # time when changes happen
            self._ndx += 1
            if self._ndx < len(self.t):
                next_t = self.t[self._ndx] - t0
            else:
                next_t = self.tmax - t0
            next_t += self._Delta_t + t0

            # if the end of the list is reached and the
            # network is supposed to be looped,
            # reset the iteration pointer and
            # and increase the time loop delta.
            if self._ndx == len(self.t) and self.loop_network:
                self._ndx = 0
                self._Delta_t += self.T

        # if the end of the list is reached and
        # the network is not supposed to be looped,
        # stop the iteration
        else:
            raise StopIteration

        return edges, t, next_t

class TemporalNetworkSimulation():
    """
    Simulation of a StochasticEpiModel on a
    temporal network.

    Parameters
    ==========
    temporal_network : TemporalNetwork
        A temporal network object.
    stochastic_epi_model : epipack.stochastic_epi_models.StochasticEpiModel
        An instance of StochasticEpiModel, initialized
        with processes and initial conditions.

    Example
    =======

    >>> model = StochasticEpiModel(["S","I","R"],3)\\
    >>>             .set_link_transmission_processes([
    >>>                     ("I", "S", 1.0, "I", "I"),
    >>>                 ])\\
    >>>             .set_node_transition_processes([
    >>>                     ("I", 2.0, "R"),
    >>>                 ])\\
    >>>             .set_node_statuses([1,0,0])
    >>> temporal_network = TemporalNetwork(3,edges,[0,0.5,1.2],1.5)
    >>> sim = TemporalNetworkSimulation(temporal_network, model)
    >>> sim.simulate(tmax=100)
    [ 0.          0.38740794  0.8210494   3.60987397  9.46213129 12.14301704] {'S': array([2., 1., 0., 0., 0., 0.]), 'I': array([1., 2., 3., 2., 1., 0.]), 'R': array([0., 0., 0., 1., 2., 3.])}

    Attributes
    ==========
    temporal_network : TemporalNetwork
        A temporal network object.
    model : epipack.stochastic_epi_models.StochasticEpiModel
        An instance of StochasticEpiModel, initialized
        with processes and initial conditions.
    """

    def __init__(self,temporal_network,stochastic_epi_model):
        self.temporal_network = temporal_network
        self.model = stochastic_epi_model
        self._initial_node_status = np.array(self.model.node_status)

        assert(self.temporal_network.N == self.model.N_nodes)

    def simulate(self,
                 tmax,
                 return_compartments=None,
                 max_unsuccessful=None,
                 sample_only_on_network_updates=False,
                 ):
        """
        Simulate a StochasticEpiModel on a temporal network.

        Parameters
        ----------
        tmax : float
            maximum length of the simulation
        return_compartments : list of compartments, default = None:
            The compartments for which to return time series.
            If ``None``, all compartments will be returned.
        max_unsuccessful : int, default = None
            The number of unsuccessful events after which the
            true total event rate will be evaluated (it might happen
            that a network becomes effectively disconnected while
            nodes are still associated with a maximum event rate).
            If ``None``, this number will be set equal to the number of nodes.
        sample_only_on_network_updates : funtion, default = None
            A function that's called when a sample is taken

        Returns
        -------
        t : numpy.ndarray
            times at which compartment counts have been sampled
        result : dict
            Dictionary mapping a compartment to a time series of its count.
        """

        if return_compartments is None:
            return_compartments = self.model.compartments

        time = [ self.temporal_network.t0() ]
        result = { C: [ 
                np.count_nonzero(
                    self.model.node_status==self.model.get_compartment_id(C)
                    )
            ] for C in return_compartments }

        for edges, t, next_t in self.temporal_network:

            if t >= tmax or self.model.simulation_has_ended():
                break

            self.model.set_network(
                    N_nodes=self.temporal_network.N,
                    edge_weight_tuples=edges,
                    directed=self.temporal_network.directed,
                    )

            _t, _res = self.model.simulate(
                    t0=t,
                    tmax=next_t,
                    return_compartments=return_compartments,
                    max_unsuccessful=max_unsuccessful,
                    stop_simulation_on_empty_network=False,
                    )

            if next_t > tmax:
                next_t == tmax

            # only save result if in fact anything changed (i.e.
            # if the number of sampling times exceeds one.
            if len(_t) > 1:
                if sample_only_on_network_updates:
                    time.append(next_t)
                    for C, timeseries in _res.items():
                        result[C].append(timeseries[-1])
                else:
                    # save all sampling events besides the first
                    time.extend(_t.tolist()[1:])
                    for C, timeseries in _res.items():
                        result[C].extend(timeseries.tolist()[1:])

        time = np.array(time)
        for C in result:
            result[C] = np.array(result[C])

        return time, result

    def reset(self,node_status=None):
        """Reset the model"""
        if node_status is not None:
            self._initial_node_status = np.array(node_status)

        self.model.set_node_statuses(self._initial_node_status)



                    
if __name__=="__main__": # pragma: no cover
    edges = [ [ (0,1) ], [ (0,1), (0,2) ], [] ]
    temporal_network = TemporalNetwork(3,edges,[0,0.5,0.6],1.0)
    for edge_list, t, next_t in temporal_network:
        if t >= 3.0:
            break
        print(t, next_t, edge_list)

    print("Delta T =",temporal_network._Delta_t)
    for edge_list, t, next_t in temporal_network:
        print("DeltaT =",temporal_network._Delta_t)
        break


    from epipack import StochasticEpiModel

    infection_rate = 1.0
    recovery_rate = 0.2
    model = StochasticEpiModel(["S","I","R"],3)\
                .set_link_transmission_processes([
                        ("I", "S", infection_rate, "I", "I"),
                    ])\
                .set_node_transition_processes([
                        ("I", recovery_rate, "R"),
                    ])\
                .set_node_statuses([1,0,0])

    temporal_network = TemporalNetwork(3,edges,[0,0.5,1.2],1.5)
    sim = TemporalNetworkSimulation(temporal_network, model)
    t, res = sim.simulate(100)
    print(t, res)
    sim.reset()
    print(sim.model.node_status)
    t, res = sim.simulate(100)
    print(t, res)

    import matplotlib.pyplot as pl
    from tqdm import tqdm

    N_meas = 10000
    #model = StochasticEpiModel(["S","I","R"],3)\
    #            .set_link_transmission_processes([
    #                    ("I", "S", infection_rate, "I", "I"),
    #                ])\
    #            .set_node_transition_processes([
    #                    ("I", recovery_rate, "R"),
    #                ])\
    model.set_node_statuses([1,0,0])

    #temporal_network = TemporalNetwork(3,edges,[0,0.5,1.5],3.0)
    #sim = TemporalNetworkSimulation(temporal_network, model)

    taus = []
    for meas in tqdm(range(N_meas)):
        sim.reset()
        t, res = sim.simulate(1000)
        if t[-1] == 0:
            continue
        else:
            taus.append(t[1])

    pl.hist(taus,bins=200,density=True)

    def rate(t):
        t = t % 1.5
        if t < 0.5:
            return infection_rate + recovery_rate
        elif t < 1.2:
            return 2*infection_rate + recovery_rate
        elif t < 1.5:
            return recovery_rate

    from scipy.integrate import cumtrapz

    t = np.linspace(0,max(taus),10000)
    rates = np.array([rate(_t) for _t in t])
    RATES = cumtrapz(rates,t,initial=0.0)
    pl.plot(t,rates*np.exp(-RATES))
    #pl.yscale('log')
    pl.show()



# -*- coding: utf-8 -*-
"""
This module provides functions related to the simulation and
measurement of epidemics.
"""

import numpy as np

from scipy.sparse import csr_matrix
from scipy.sparse import eye
from scipy.linalg import expm
from scipy.sparse.linalg import eigs
from scipy.optimize import minimize


_EPI_S = 0
_EPI_I = 1
_EPI_R = 2

class SIR_weighted:

    infected = []
    force_of_infection_on_node = []
    node_status = []
    rates = np.array([0.0,0.0])

    def __init__(self,N,weighted_edge_tuples,infection_rate,recovery_rate,number_of_initially_infected):

        self.N = N
        self.infection_rate = infection_rate
        self.recovery_rate = recovery_rate
        self.infected = np.random.choice(N,replace=False,size=(number_of_initially_infected,)).tolist()
        self.recovered = []
        self.node_status = np.array([ _EPI_S for n in range(N) ],dtype=int)
        for n in self.infected:
            self.node_status[n] = _EPI_I 


        self.force_of_infection_on_node = np.array([ 0.0 for n in range(N) ])
        self.graph = [ {} for n in range(N) ]
        for u, v, weight in weighted_edge_tuples:
            self.graph[u][v] = weight
            self.graph[v][u] = weight
            if self.node_status[u] == _EPI_S and self.node_status[v] == _EPI_I:
                self.force_of_infection_on_node[u] += weight
            elif self.node_status[v] == _EPI_S and self.node_status[u] == _EPI_I:
                self.force_of_infection_on_node[v] += weight

    def evaluate_rates(self):
        self.rates[0] = self.force_of_infection_on_node.sum() * self.infection_rate
        self.rates[1] = len(self.infected) * self.recovery_rate
        return self.rates

    def recovery_event(self):
        recovering_node = self.infected.pop( np.random.choice(len(self.infected)) )
        self.node_status[recovering_node] = _EPI_R

        self.force_of_infection_on_node[recovering_node] = 0.0
        
        for neigh, weight in self.graph[recovering_node].items():
            if self.node_status[neigh] == _EPI_S:
                self.force_of_infection_on_node[neigh] -= weight
                if self.force_of_infection_on_node[neigh] < 0.0:
                    self.force_of_infection_on_node[neigh] = 0.0

        self.recovered.append(recovering_node)

    def infection_event(self):
        infecting_node = np.random.choice(self.N, 
                                          p=self.force_of_infection_on_node/self.force_of_infection_on_node.sum()
                                         )
        self.node_status[infecting_node] = _EPI_I
        self.force_of_infection_on_node[infecting_node] = 0.0

        for neigh, weight in self.graph[infecting_node].items():
            if self.node_status[neigh] == _EPI_S:
                self.force_of_infection_on_node[neigh] += weight

        self.infected.append(infecting_node)

    def simulation(self,tmax):

        t = 0.0
        time = []
        I = []
        R = []

        while (t < tmax) and len(self.infected) > 0:

            time.append(t)
            I.append(len(self.infected))
            R.append(len(self.recovered))

            r = self.evaluate_rates()
            Lambda = r.sum()

            tau = np.random.exponential(1.0/Lambda)
            event = np.random.choice(len(r),p=r/Lambda)

            if event == 0:
                self.infection_event()
            else:
                self.recovery_event()

            t += tau

        if len(self.infected) == 0 and (t < tmax):
            time.append(t)
            I.append(0)
            R.append(len(self.recovered))


        return np.array(time), np.array(I, dtype=float), np.array(R, dtype=float)


if __name__ == "__main__":

    N = 100
    k = 10
    omega = 1.6
    recovery_rate = 0.1
    R0 = 10
    t_run_total = 1000




import unittest

import numpy as np
import sympy

from epipack.networks import (
            get_2D_lattice_links
        )

_xor = lambda a, b: (a and not b) or (not a and b)

class NetworkModelsTest(unittest.TestCase):

    def test_unperiodic_undiagonal(self):

        links = get_2D_lattice_links(2)

        expected_links = [
                    (0,1,1.0),
                    (0,2,1.0),
                    (1,3,1.0),
                    (2,3,1.0),
                ]

        assert(set(links) == set(expected_links))

    def test_periodic_undiagonal(self):

        links = get_2D_lattice_links(2,periodic=True)

        expected_links = [
                    (0,1,1.0),
                    (0,2,1.0),
                    (1,3,1.0),
                    (2,3,1.0),
                ]

        assert(set(links) == set(expected_links))

    def test_unperiodic_diagonal(self):

        links = get_2D_lattice_links(2,diagonal_links=True)

        expected_links = [
                    (0,1,1.0),
                    (0,2,1.0),
                    (0,3,1.0),
                    (1,3,1.0),
                    (1,2,1.0),
                    (2,3,1.0),
                ]

        assert(set(links) == set(expected_links))

    def test_periodic_diagonal(self):

        links = get_2D_lattice_links(2,periodic=True,diagonal_links=True)

        expected_links = [
                    (0,1,1.0),
                    (0,2,1.0),
                    (0,3,1.0),
                    (1,3,1.0),
                    (1,2,1.0),
                    (2,3,1.0),
                ]

        assert(set(links) == set(expected_links))

if __name__ == "__main__":

    T = NetworkModelsTest()
    T.test_unperiodic_undiagonal()
    T.test_unperiodic_diagonal()
    T.test_periodic_undiagonal()
    T.test_periodic_diagonal()



"""
Provides an API to define Matrix epidemiological models.
"""

import numpy as np
import scipy.sparse as sprs

import warnings

from epipack.integrators import (
            integrate_dopri5,
            integrate_euler,
            IntegrationMixin,
        )

from epipack.process_conversions import (
            processes_to_rates,
            transition_processes_to_rates,
            fission_processes_to_rates,
            fusion_processes_to_rates,
            transmission_processes_to_rates,
        )


class MatrixEpiModel(IntegrationMixin):
    """
    A general class to define standard
    mean-field compartmental
    epidemiological model.

    Parameters
    ----------

    compartments : :obj:`list` of :obj:`string`
        A list containing compartment strings.

    Attributes
    ----------

    compartments : :obj:`list` of :obj:`string`
        A list containing strings that describe each compartment,
        (e.g. "S", "I", etc.).
    N_comp : :obj:`int`
        Number of compartments (including population number)
    linear_rates : scipy.sparse.csr_matrix
        Sparse matrix containing
        transition rates of the linear processes.
    quadratic_rates : scipy.sparse.csr_matrix
        List of sparse matrices that contain
        transition rates of the quadratic processes
        for each compartment.
    affected_by_quadratic_process : :obj:`list` of :obj:`int`
        List of integer compartment IDs, collecting
        compartments that are affected
        by the quadratic processes


    Example
    -------

    .. code:: python

        >>> epi = MatrixEpiModel(["S","I","R"])
        >>> print(epi.compartments)
        [ "S", "I", "R" ]


    """

    def __init__(self,compartments,initial_population_size=1,correct_for_dynamical_population_size=False):
        """
        """

        self.t0 = None
        self.y0 = None
        self.affected_by_quadratic_process = []

        self.compartments = list(compartments)
        self.compartment_ids = { C: iC for iC, C in enumerate(compartments) }
        self.initial_population_size = initial_population_size
        self.correct_for_dynamical_population_size = correct_for_dynamical_population_size
        self.N_comp = len(self.compartments)
        self.birth_rates = np.zeros((self.N_comp,),dtype=np.float64)
        self.linear_rates = sprs.csr_matrix((self.N_comp, self.N_comp),dtype=np.float64)
        self.quadratic_rates = [ sprs.csr_matrix((self.N_comp, self.N_comp),dtype=np.float64)\
                                 for c in range(self.N_comp) ]

    def get_compartment_id(self,C):
        """Get the integer ID of a compartment ``C``"""
        return self.compartment_ids[C]

    def get_compartment(self,iC):
        """Get the compartment, given an integer ID ``iC``"""
        return self.compartments[iC]

    def set_processes(self,
                      process_list,
                      allow_nonzero_column_sums=False,
                      reset_rates=True,
                      ignore_rate_position_checks=False,
                      ):
        """
        Converts a list of reaction process tuples to rate tuples and sets the rates for this model.

        Parameters
        ----------
        process_list : :obj:`list` of :obj:`tuple`
            A list containing reaction processes in terms of tuples.

            .. code:: python

                [
                    # transition process
                    ( source_compartment, rate, target_compartment),

                    # transmission process
                    ( coupling_compartment_0, coupling_compartment_1, rate, target_compartment_0, target_ccompartment_1),

                    # fission process
                    ( source_compartment, rate, target_compartment_0, target_ccompartment_1),

                    # fusion process
                    ( source_compartment_0, source_compartment_1, rate, target_compartment),

                    # death process
                    ( source_compartment, rate, None),

                    # birth process
                    ( None, rate, target_compartment),
                ]
        allow_nonzero_column_sums : bool, default : False
            Traditionally, epidemiological models preserve the
            total population size. If that's not the case,
            switch off testing for this.
        reset_rates : bool, default : True
            If this is `True`, reset all rates to zero before setting the new ones.
        ignore_rate_position_checks : bool, default = False
            This function usually checks whether the rate of
            a reaction is positioned correctly. You can
            turn this behavior off for transition, birth, death, and
            transmission processes. (Useful if you want to define
            symbolic transmission processes that are compartment-dependent).
        """

        quadratic_rates, linear_rates = processes_to_rates(process_list,
                                                           self.compartments,
                                                           ignore_rate_position_checks,
                                                           )
        self.set_linear_rates(linear_rates,allow_nonzero_column_sums=allow_nonzero_column_sums)
        self.set_quadratic_rates(quadratic_rates,allow_nonzero_column_sums=allow_nonzero_column_sums)

        return self

    def set_linear_rates(self,rate_list,allow_nonzero_column_sums=False,reset_rates=True):
        """
        Define the linear transition rates between compartments.

        Parameters
        ==========

        rate_list : :obj:`list` of :obj:`tuple`
            A list of tuples that contains transitions rates in the following format:

            .. code:: python

                [
                    ( acting_compartment, affected_compartment, rate ),
                    ...
                ]

        allow_nonzero_column_sums : :obj:`bool`, default : False
            Traditionally, epidemiological models preserve the
            total population size. If that's not the case,
            switch off testing for this.

        Example

        reset_rates : bool, default : True
            Whether to reset all linear rates to zero before
            converting those.
        """

        if reset_rates:
            row = []
            col = []
            data = []
            birth_rates = np.zeros((self.N_comp,),dtype=np.float64)
        else:
            linear_rates = self.linear_rates.tocoo()
            row = linear_rates.row.tolist()
            col = linear_rates.col.tolist()
            data = linear_rates.data.tolist()
            birth_rates = self.birth_rates.copy()

        for acting_compartment, affected_compartment, rate in rate_list:

            _t = self.get_compartment_id(affected_compartment)
            if acting_compartment is None:
                birth_rates[_t] += rate
            else:
                _s = self.get_compartment_id(acting_compartment)
                row.append(_t)
                col.append(_s)
                data.append(rate)

        linear_rates = sprs.coo_matrix((data, (row, col)),shape=(self.N_comp, self.N_comp),dtype=np.float64)
        linear_rates = linear_rates.tocsr()

        if not allow_nonzero_column_sums:
            test = linear_rates.sum(axis=0) + birth_rates
            test_sum = test.sum()
            if np.abs(test_sum) > 1e-15:
                warnings.warn("Rates do not sum to zero for each column:" + str(test_sum))

        self.linear_rates = linear_rates
        self.birth_rates = birth_rates

        return self


    def add_transition_processes(self,process_list):
        """
        Define the linear transition processes between compartments.

        Parameters
        ==========

        process_list : :obj:`list` of :obj:`tuple`
            A list of tuples that contains transitions rates in the following format:

            .. code:: python

                [
                    ( source_compartment, rate, target_compartment ),
                    ...
                ]

        Example
        -------

        For an SEIR model.

        .. code:: python

            epi.add_transition_processes([
                ("E", symptomatic_rate, "I" ),
                ("I", recovery_rate, "R" ),
            ])

        """

        linear_rates = transition_processes_to_rates(process_list)

        return self.set_linear_rates(linear_rates,
                                     reset_rates=False,
                                     allow_nonzero_column_sums=True,
                                     )

    def add_fission_processes(self,process_list):
        """
        Define linear fission processes between compartments.

        Parameters
        ==========

        process_list : :obj:`list` of :obj:`tuple`
            A list of tuples that contains fission rates in the following format:

            .. code:: python

                [
                    ("source_compartment", rate, "target_compartment_0", "target_compartment_1" ),
                    ...
                ]

        Example
        -------

        For pure exponential growth of compartment `B`.

        .. code:: python

            epi.add_fission_processes([
                ("B", growth_rate, "B", "B" ),
            ])

        """
        linear_rates = fission_processes_to_rates(process_list)

        return self.set_linear_rates(linear_rates,
                                     reset_rates=False,
                                     allow_nonzero_column_sums=True,
                                     )

    def add_fusion_processes(self,process_list):
        """
        Define fusion processes between compartments.

        Parameters
        ==========

        process_list : :obj:`list` of :obj:`tuple`
            A list of tuples that contains fission rates in the following format:

            .. code:: python

                [
                    ("coupling_compartment_0", "coupling_compartment_1", rate, "target_compartment_0" ),
                    ...
                ]

        Example
        -------

        Fusion of reactants "A", and "B" to form "C".

        .. code:: python

            epi.add_fusion_processes([
                ("A", "B", reaction_rate, "C" ),
            ])

        """
        quadratic_rates = fusion_processes_to_rates(process_list)

        return self.set_quadratic_rates(quadratic_rates,
                                        reset_rates=False,
                                        allow_nonzero_column_sums=True
                                        )

    def add_transmission_processes(self,process_list):
        r"""
        A wrapper to define quadratic process rates through
        transmission reaction equations. Note that in stochastic
        network/agent simulations, the transmission
        rate is equal to a rate per link. For the mean-field ODEs,
        the rates provided to this function will just be equal
        to the prefactor of the respective quadratic terms.

        For instance, if you analyze an SIR system and simulate on a network of mean degree :math:`k_0`,
        a basic reproduction number :math:`R_0`, and a recovery rate :math:`\mu`,
        you would define the single link transmission process as

            .. code:: python

                ("I", "S", R_0/k_0 * mu, "I", "I")

        For the mean-field system here, the corresponding reaction equation would read

            .. code:: python

                ("I", "S", R_0 * mu, "I", "I")

        Parameters
        ----------
        process_list : :obj:`list` of :obj:`tuple`
            A list of tuples that contains transitions rates in the following format:

            .. code:: python

                [
                    ("source_compartment",
                     "target_compartment_initial",
                     rate,
                     "source_compartment",
                     "target_compartment_final",
                     ),
                    ...
                ]

        Example
        -------

        For an SEIR model.

        .. code:: python

            epi.add_transmission_processes([
                ("I", "S", +1, "I", "E" ),
            ])

        """
        quadratic_rates = transmission_processes_to_rates(process_list)

        return self.set_quadratic_rates(quadratic_rates,
                                        reset_rates=False,
                                        allow_nonzero_column_sums=True,
                                        )

    def add_quadratic_rates(self,rate_list,reset_rates=True,allow_nonzero_column_sums=False):
        """
        Add quadratic rates without resetting the existing rate terms.
        See :func:`_tacoma.set_quadratic_rates` for docstring.
        """

        return self.set_quadratic_rates(rate_list,
                                        reset_rates=False,
                                        allow_nonzero_column_sums=allow_nonzero_column_sums,
                                        )

    def add_linear_rates(self,rate_list,reset_rates=True,allow_nonzero_column_sums=False):
        """
        Add linear rates without resetting the existing rate terms.
        See :func:`_tacoma.set_linear_rates` for docstring.
        """

        return self.set_linear_rates(rate_list,
                                     reset_rates=False,
                                     allow_nonzero_column_sums=allow_nonzero_column_sums,
                                     )

    def set_quadratic_rates(self,rate_list,reset_rates=True,allow_nonzero_column_sums=False):
        r"""
        Define the quadratic transition processes between compartments.

        Parameters
        ----------
        rate_list : :obj:`list` of :obj:`tuple`
            A list of tuples that contains transitions rates in the following format:

            .. code:: python

                [
                    (
                      "coupling_compartment_0",
                      "coupling_compartment_1",
                      "affected_compartment",
                      rate
                     ),
                    ...
                ]

        allow_nonzero_column_sums : :obj:`bool`, default : False
            Traditionally, epidemiological models preserve the
            total population size. If that's not the case,
            switch off testing for this.

        Example
        -------

        For an SEIR model.

        .. code:: python

            epi.set_quadratic_rates([
                ("S", "I", "S", -1 ),
                ("S", "I", "E", +1 )
            ])

        Read  as

        "Coupling of *S* and *I* leads to
        a reduction in "S" proportional to :math:`S\times I` and rate -1/time_unit.
        Furthermore, coupling of *S* and *I* leads to
        an increase in "E" proportional to :math:`S\times I` and rate +1/time_unit."
        """

        matrices = [ [[], [], []] for c in self.compartments]
        if reset_rates:
            all_affected = []
        else:
            for iC, _M in enumerate(self.quadratic_rates):
                if _M.count_nonzero() > 0:
                    M = _M.tocoo()
                    matrices[iC][0] = M.row.tolist()
                    matrices[iC][1] = M.col.tolist()
                    matrices[iC][2] = M.data.tolist()
                else:
                    matrices[iC] = [ [], [], [] ]
            all_affected = self.affected_by_quadratic_process if len(self.affected_by_quadratic_process)>0 else []

        for coupling0, coupling1, affected, rate in rate_list:

            c0, c1 = sorted([ self.get_compartment_id(c) for c in [coupling0, coupling1] ])
            a = self.get_compartment_id(affected)

            matrices[a][0].append(c0)
            matrices[a][1].append(c1)
            matrices[a][2].append(rate)
            all_affected.append(a)

        total_sum = 0
        for c in range(self.N_comp):
            row = matrices[c][0]
            col = matrices[c][1]
            data = matrices[c][2]
            if len(row) > 0:
                matrices[c] = sprs.coo_matrix((data,(row, col)),
                                               shape=(self.N_comp, self.N_comp),
                                               dtype=np.float64,
                                              ).tocsr()
                total_sum += matrices[c].sum(axis=0).sum()
            else:
                matrices[c] = sprs.csr_matrix((self.N_comp, self.N_comp),dtype=np.float64)

        if np.abs(total_sum) > 1e-14:
            if not allow_nonzero_column_sums:
                warnings.warn("Rates do not sum to zero. Sum = "+ str(total_sum))

        self.affected_by_quadratic_process = sorted(list(set(all_affected)))
        self.quadratic_rates = matrices

        return self

    def dydt(self,t,y):
        """
        Compute the current momenta of the epidemiological model.

        Parameters
        ----------
        t : :obj:`float`
            Current time
        y : numpy.ndarray
            The first entry is equal to the population size.
            The remaining entries are equal to the current fractions
            of the population of the respective compartments
        """

        ynew = self.linear_rates.dot(y) + self.birth_rates
        if self.correct_for_dynamical_population_size:
            population_size = y.sum()
        else:
            population_size = self.initial_population_size
        for c in self.affected_by_quadratic_process:
            ynew[c] += y.T.dot(self.quadratic_rates[c].dot(y)) / population_size

        return ynew

    def get_numerical_dydt(self):
        """
        Return a function that obtains t and y as an input and returns dydt of this system
        """
        return self.dydt

    def jacobian(self,y0=None):
        """
        Return Jacobian at point ``y0``. Will use ``self.y0`` if argument ``y0`` is ``None``.
        """
        return self.get_transmission_matrix(y0=y0) + self.get_transition_matrix()

    def get_jacobian_leading_eigenvalue(self,y0=None,returntype='complex'):
        """
        Return leading eigenvalue of Jacobian at point ``y0``.
        Will use ``self.y0`` if argument ``y0`` is ``None``.
        Use argument ``returntype='real'`` to only obtain the
        real part of the eigenvalue.
        """
        J = self.jacobian(y0=y0)
        return self._get_leading_eigenvalue(J,returntype,'LR')

    def get_transmission_matrix(self,y0=None):
        """
        Return transmission matrix at point ``y0``.
        Will use ``self.y0`` if argument ``y0`` is ``None``.
        """
        if y0 is None:
            y0 = self.y0

        if y0 is None:
            raise ValueError("No initial conditions have been given or set.")

        if self.correct_for_dynamical_population_size:
            raise NotImplementedError("transmission matrices for models with "
                                      "varying population size have not been implemented.")

        rows = []
        for M in self.quadratic_rates:
            row = ((M+M.T).dot(y0)) / self.initial_population_size
            row = row.reshape(1,self.N_comp)
            row = sprs.csr_matrix(row)
            rows.append(row)

        T = sprs.vstack(rows,format='csr')

        return T

    def get_transition_matrix(self):
        """
        Return transition matrix.
        """
        return self.linear_rates

    def get_next_generation_matrix(self,y0=None):
        """
        Return next generation matrix at point y0.
        Will use ``self.y0`` if argument ``y0`` is ``None``.
        """
        T = self.get_transmission_matrix(y0=y0)
        Sigma = self.get_transition_matrix()

        # delete all compartments that contribute to the singularity
        # of the transition matrix
        del_cols = []
        del_rows = []

        n = Sigma.shape[0]

        for i in range(n):
            if np.all(Sigma[i,:].data==0):
                del_rows.append(i)
        for j in range(n):
            if np.all(Sigma[:,j].data==0):
                del_cols.append(j)

        del_comp = set(del_cols) | set(del_rows)
        use_comp = set(range(n)) - del_comp

        use_comp = sorted(list(use_comp))

        # filter our compartments that do not make the matrix singular
        Sigma = (Sigma[use_comp,:])[:,use_comp]
        T = (T[use_comp,:])[:,use_comp]

        # convert Sigma to csc for more efficient inverse algo
        K = -T.dot(sprs.linalg.inv(Sigma.tocsc()))

        return K

    def get_next_generation_matrix_leading_eigenvalue(self,y0=None,returntype='real'):
        """
        Return the leading eigenvalue of the next generation matrix
        at point ``y0``. Will use ``self.y0`` if argument
        ``y0`` is ``None``. When ``y0`` is equal to the initial
        disease-free state, this function will return the
        basic reproduction number :math:`R_0`.

        The function will return only the real part by default.
        Use ``returntype='complex'`` to change this.
        """
        K = self.get_next_generation_matrix(y0=y0)
        return self._get_leading_eigenvalue(K,returntype,method='LM')

    def _get_leading_eigenvalue(self,M,returntype='complex',method='LR'):

        if M.shape == (1,1):
            _lambda = M[0,0]
        elif M.shape == (1,):
            _lambda = M[0]
        else:
            # I thought CSC format would be better for solver
            # but it's not, so I uncommented this
            # M_ = M.tocsc()
            M_ = M
            if M_.shape == (2,2):
                lambdas = np.linalg.eig(M_.toarray())[0]
            else:
                lambdas = sprs.linalg.eigs(M_,k=min(2,M_.shape[0]-2),which=method)[0]
            if method == 'LR':
                lambdas = sorted(lambdas, key=lambda x: -np.real(x))
                _lambda = lambdas[0]
                if returntype == 'real':
                    _lambda = np.real(_lambda)
            elif method == 'LM':
                lambdas = sorted(lambdas, key=lambda x: -np.abs(x))
                _lambda = lambdas[0]
                if returntype == 'real':
                    _lambda = np.abs(_lambda)
        return _lambda


class MatrixSIModel(MatrixEpiModel):
    """
    An SI model derived from :class:`epipack.numeric_matrix_based_epi_models.MatrixEpiModel`.
    """

    def __init__(self, infection_rate, initial_population_size=1.0):

        MatrixEpiModel.__init__(self, list("SI"), initial_population_size)

        self.set_quadratic_rates([
                ("S", "I", "S", -infection_rate),
                ("S", "I", "I", +infection_rate),
            ])


class MatrixSISModel(MatrixEpiModel):
    """
    An SIS model derived from :class:`epipack.numeric_matrix_based_epi_models.MatrixEpiModel`.

    Parameters
    ----------
    R0 : float
        The basic reproduction number. From this, the infection
        rate is computed as ``infection_rate = R0 * recovery_rate``
    recovery_rate : float
        Recovery rate
    population_size : float, default = 1.0
        Number of people in the population.
    """

    def __init__(self, R0, recovery_rate, initial_population_size=1.0):

        infection_rate = R0 * recovery_rate

        MatrixEpiModel.__init__(self, list("SI"), initial_population_size)

        self.set_quadratic_rates([
                ("S", "I", "S", -infection_rate),
                ("S", "I", "I", +infection_rate),
            ])
        self.add_transition_processes([
                ("I", recovery_rate, "S" ),
            ])

class MatrixSIRModel(MatrixEpiModel):
    """
    An SIR model derived from :class:`epipack.numeric_matrix_based_epi_models.MatrixEpiModel`.
    """

    def __init__(self, R0, recovery_rate, initial_population_size=1.0):

        infection_rate = R0 * recovery_rate

        MatrixEpiModel.__init__(self, list("SIR"), initial_population_size)

        self.set_quadratic_rates([
                ("S", "I", "S", -infection_rate),
                ("S", "I", "I", +infection_rate),
            ])
        self.add_transition_processes([
                ("I", recovery_rate, "R"),
            ])

class MatrixSIRSModel(MatrixEpiModel):
    """
    An SIRS model derived from :class:`epipack.numeric_matrix_based_epi_models.MatrixEpiModel`.
    """

    def __init__(self, R0, recovery_rate, waning_immunity_rate, initial_population_size=1.0):

        infection_rate = R0 * recovery_rate

        MatrixEpiModel.__init__(self, list("SIR"), initial_population_size)

        self.set_quadratic_rates([
                ("S", "I", "S", -infection_rate),
                ("S", "I", "I", +infection_rate),
            ])
        self.add_transition_processes([
                ("I", recovery_rate, "R"),
                ("R", waning_immunity_rate, "S"),
            ])

class MatrixSEIRModel(MatrixEpiModel):
    """
    An SEIR model derived from :class:`epipack.numeric_matrix_based_epi_models.MatrixEpiModel`.
    """

    def __init__(self, R0, recovery_rate, symptomatic_rate, initial_population_size=1.0):

        infection_rate = R0 * recovery_rate

        MatrixEpiModel.__init__(self, list("SEIR"), initial_population_size)

        self.set_quadratic_rates([
                ("S", "I", "S", -infection_rate),
                ("S", "I", "E", +infection_rate),
            ])
        self.add_transition_processes([
                ("E", symptomatic_rate, "I"),
                ("I", recovery_rate, "R"),
            ])


if __name__=="__main__":    # pragma: no cover
    epi = MatrixEpiModel(list("SEIR"))
    print(epi.compartments)
    print()
    epi.add_transition_processes([
            ("E", 1.0, "I"),
            ("I", 1.0, "R"),
            ])
    print(epi.linear_rates)
    epi.set_quadratic_rates([
            ("S", "I", "S", -1.0),
            ("S", "I", "E", +1.0),
            ])
    print()
    for iM, M in enumerate(epi.quadratic_rates):
        print(epi.get_compartment(iM), M)

    import matplotlib.pyplot as pl

    N = 100
    epi = MatrixSISModel(R0=2,recovery_rate=1,initial_population_size=N)
    print(epi.linear_rates)
    epi.set_initial_conditions({'S':0.99*N,'I':0.01*N})
    tt = np.linspace(0,10,100)
    result = epi.integrate(tt,['S','I'])

    pl.plot(tt, result['S'],label='S')
    pl.plot(tt, result['I'],label='I')
    pl.legend()

    pl.show()
